"""
The GAMMA visualization library, providing `model/view/controller` oriented classes for
rendering data in different shapes, styles and formats, e.g., as matplot charts or
plain text.
"""
import logging
from abc import ABCMeta, abstractmethod
from threading import Lock
from typing import Any, Dict, Generic, Iterable, Optional, Type, TypeVar, Union, cast

from ..api import AllTracker, inheritdoc
from .color import ColorScheme, FacetDarkColorScheme, FacetLightColorScheme

log = logging.getLogger(__name__)

#
# Exported names
#

__all__ = ["DrawingStyle", "ColoredStyle", "Drawer"]

#
# Type variables
#

T = TypeVar("T")
T_Model = TypeVar("T_Model")
T_Style = TypeVar("T_Style", bound="DrawingStyle")
T_Style_Class = TypeVar("T_Style_Class", bound=Type["DrawingStyle"])
T_ColorScheme = TypeVar("T_ColorScheme", bound=ColorScheme)


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


# View: class DrawingStyle


class DrawingStyle(metaclass=ABCMeta):
    """
    Base class for styles used by :class:`.Drawer` objects.

    Provides low-level rendering methods used by specific type of :class:`.Drawer`.
    Typically, there are several drawing styles for one drawer type, offering the same
    rendering methods but implementing them differently (e.g., matplot output vs. text
    output).
    The style class and its rendering methods should not be aware of the actual object
    to be rendered; overall control of the rendering process stays with the
    :class:`.Drawer`; the associated style object only carries out the low-level
    rendering and controls formatting.

    For example, a :class:`.MatrixDrawer` requires a :class:`.MatrixStyle` to render
    matrices.
    :class:`.MatrixStyle` is an abstract subclass of :class:`DrawingStyle` and has three
    implementations to output matrices as matplot charts or as a text report:
    :class:`.MatrixMatplotStyle`, :class:`.PercentageMatrixMatplotStyle`,
    and :class:`.MatrixReportStyle`.

    Many style objects can be further parameterised to control how objects are rendered.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if len(kwargs) > 0:
            raise KeyError(
                f'unknown argument{"s" if len(kwargs) > 1 else ""}: '
                f'{", ".join(kwargs.keys())}'
            )
        self._lock = Lock()

    @classmethod
    def get_named_styles(cls: T_Style_Class) -> Dict[str, T_Style_Class]:
        """
        Get a mapping of names to default instances of this style class.

        :return: a dictionary mapping of names to default instances of this style class
        """
        return {cls.get_default_style_name(): cls}

    @classmethod
    @abstractmethod
    def get_default_style_name(cls) -> str:
        """
        Get the name of the default style associated with this style class.

        The default style is obtained by instantiating this style class without
        parameters.

        Common examples for default style names are `matplot` and `text`.

        :return: the name of the default style
        """

    @abstractmethod
    def start_drawing(self, title: str, **kwargs) -> None:
        """
        Prepare a new chart for drawing, using the given title.

        Any additional drawer-specific attributes, obtained from
        method :meth:`Drawer._get_style_kwargs`, will be passed
        as keyword arguments.

        :param title: the title of the chart
        """

    def finalize_drawing(self, **kwargs) -> None:
        """
        Finalize the drawing.

        Any additional drawer-specific attributes, obtained from
        method :meth:`.Drawer._get_style_kwargs`, will be passed
        as keyword arguments.
        """


@inheritdoc(match="[see superclass]")
class ColoredStyle(DrawingStyle, Generic[T_ColorScheme], metaclass=ABCMeta):
    """
    A drawing style that supports color output.
    """

    def __init__(self, *, colors: Optional[T_ColorScheme] = None, **kwargs) -> None:
        """
        :param colors: the color scheme to be used by this drawing style
        """
        super().__init__(**kwargs)
        self._colors = colors or FacetLightColorScheme()

    @classmethod
    def dark(cls: T_Style_Class) -> T_Style_Class:
        """
        Create a dark variant of the given drawing style, using the default dark
        background color scheme :class:`.FacetDarkColorScheme`.

        :return: the dark drawing style class
        """

        class DarkStyle(ColoredStyle, metaclass=ABCMeta):
            """
            The dark variant of a :class:`.ColoredStyle`, using the
            :class:`.FacetDarkColorScheme`.
            """

            def __init__(self, **kwargs) -> None:
                """
                :param kwargs: init parameters for the new drawing style instance
                """
                super().__init__(colors=FacetDarkColorScheme(), **kwargs)

        dark_style_class = cast(
            T_Style_Class, type(f"Dark{cls.__name__}", (DarkStyle, cls), {})
        )
        dark_style_class.__module__ = cls.__module__
        return dark_style_class

    @classmethod
    def get_named_styles(cls: T_Style_Class) -> Dict[str, T_Style_Class]:
        """[see superclass]"""
        named_styles = super().get_named_styles()
        return {
            **named_styles,
            **{
                f"{name}_dark": cast(ColoredStyle, style).dark()
                for name, style in named_styles.items()
                if isinstance(style, type)
            },
        }

    @property
    def colors(self) -> T_ColorScheme:
        """
        The color scheme used by this style.
        """
        return self._colors


# Controller: class Drawer


class Drawer(Generic[T_Model, T_Style], metaclass=ABCMeta):
    """
    Base class for drawers.

    Drawers follow a `Model-View-Controller` design `(MVC)`.
    Each :class:`Drawer` (the `controller`) is associated with a :class:`.DrawingStyle`
    object (the `view`) and renders objects of a specific type (the `model`),
    using the low-level drawing methods provided by the style object.

    While the drawer controls the overall drawing process (e.g., drawing a tree or a
    matrix), the style objects determines the format of the output, e.g., a text or
    a line drawing.
    """

    #: The :class:`.DrawingStyle` used by this drawer.
    style: T_Style

    #: The name of the default drawing style.
    DEFAULT_STYLE = "matplot"

    def __init__(self, style: Optional[Union[T_Style, str]] = None) -> None:
        """
        :param style: the style to be used for drawing; either as a
            :class:`.DrawingStyle` instance, or as the name of a default style.
            Permissible names include ``"matplot"`` for a style supporting `matplotlib`,
            and ``"text"`` if text rendering is supported (default: ``"%DEFAULT%"``)
        """

        def _get_style_factory(_style_name) -> Type[T_Style]:
            # get the named style from the style dict
            try:
                return self.get_named_styles()[_style_name]
            except KeyError:
                raise KeyError(f"unknown named style: {_style_name}")

        if style is None:
            self.style = _get_style_factory(Drawer.DEFAULT_STYLE)()
        elif isinstance(style, str):
            self.style = _get_style_factory(style)()
        elif isinstance(style, DrawingStyle):
            self.style = style
        else:
            raise TypeError(
                "arg style expected to be a string, or an instance of class "
                f"{DrawingStyle.__name__}"
            )

    __init__.__doc__ = __init__.__doc__.replace("%DEFAULT%", DEFAULT_STYLE)

    @classmethod
    def get_named_styles(cls) -> Dict[str, Type[T_Style]]:
        """
        Get a mapping of names to style factories for all named styles recognized by
        this drawer's initializer.

        A factory is a class or function with no mandatory parameters.
        """

        return {
            name: style
            for style_class in cls.get_style_classes()
            for name, style in style_class.get_named_styles().items()
        }

    @classmethod
    @abstractmethod
    def get_style_classes(cls) -> Iterable[Type[T_Style]]:
        """
        Get all style classes available for this drawer type.

        :returns: an iterable of style classes
        """
        pass

    def draw(self, data: T_Model, title: str) -> None:
        """
        Render the data using the style associated with this drawer.

        :param data: the data to be rendered
        :param title: the title of the resulting chart
        """
        style = self.style

        # styles might hold some drawing context, so make sure we are thread safe
        # noinspection PyProtectedMember
        with style._lock:
            style_attributes = self._get_style_kwargs(data)
            # noinspection PyProtectedMember
            style.start_drawing(title, **style_attributes)
            self._draw(data)
            # noinspection PyProtectedMember
            style.finalize_drawing(**style_attributes)

    def _get_style_kwargs(self, data: T_Model) -> Dict[str, Any]:
        """
        Using the given data object, derive keyword arguments to be passed to the
        style's :meth:`~DrawingStyle.start_drawing` and
        :meth:`~DrawingStyle.finalize_drawing` methods.

        :meta public:
        :param data: the data to be rendered
        :returns: the style attributes for the given data object
        """
        return dict()

    @abstractmethod
    def _draw(self, data: T_Model) -> None:
        """
        Core drawing method invoked my method :meth:`.draw`.

        :meta public:
        :param data: the data to be rendered
        """
        pass


__tracker.validate()



"""
Core implementation of :mod:`pytools.meta`.
"""
import logging
from typing import Optional, Type, TypeVar
from weakref import ref

from ..api import AllTracker

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = ["SingletonMeta", "compose_meta"]


#
# Type variables
#

T = TypeVar("T")


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class SingletonMeta(type):
    """
    Meta-class for singleton classes.

    Subsequent instantiations of a singleton class return the identical object.
    Singleton classes may not accept any parameters upon instantiation.
    """

    def __init__(cls, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        cls.__instance_ref: Optional[ref] = None

    def __call__(cls: Type[T], *args, **kwargs) -> T:
        """
        Return the existing singleton instance, or create a new one if none exists yet.

        Behind the scenes, uses a weak reference so the singleton instance can be
        garbage collected when no longer in use.

        Singletons must be instantiated without any parameters.

        :return: the singleton instance
        :raises ValueError: if called with parameters
        """
        if args or kwargs:
            raise ValueError("singleton classes may not take any arguments")

        cls: SingletonMeta

        if cls.__instance_ref:
            obj = cls.__instance_ref()
            if obj is not None:
                return obj

        instance = super(SingletonMeta, cls).__call__()
        cls.__instance_ref = ref(instance)
        return instance


#
# Functions
#


def compose_meta(*metaclasses: type) -> type:
    """
    Compose multiple metaclasses.

    :param metaclasses: one or more metaclasses
    """
    metaclasses = tuple(
        mcs
        for i, mcs in enumerate(metaclasses)
        if mcs is not type and mcs not in metaclasses[:i]
    )
    return type("_" + "_".join(mcs.__name__ for mcs in metaclasses), metaclasses, {})


__tracker.validate()



#!/usr/bin/env python3
"""
Make sphinx documentation using the makefile in pytools
"""

import os
import sys


def make() -> None:
    """
    Run the common make file available in the pytools repo
    """
    cwd = os.path.dirname(os.path.realpath(__file__))
    os.chdir(cwd)

    sys.path.insert(
        0,
        os.path.normpath(
            os.path.join(cwd, os.pardir, os.pardir, "pytools", "sphinx", "base")
        ),
    )
    # noinspection PyUnresolvedReferences
    from make_base import make

    make(modules=["pytools"])


if __name__ == "__main__":
    make()



"""
Implementation of :mod:`pytools.expression` and subpackages.
"""
import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Iterable, Optional, Tuple, TypeVar

from ..api import AllTracker, inheritdoc, to_list
from .operator import BinaryOperator, UnaryOperator

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = [
    "ExpressionFormatter",
    "HasExpressionRepr",
    "Expression",
    "FrozenExpression",
    "make_expression",
    "freeze",
    "ExpressionAlias",
]

#
# Type variables
#

T = TypeVar("T")


#
# Ensure all symbols introduced below are included in __all__
#


__tracker = AllTracker(globals())

#
# Class definitions
#


class ExpressionFormatter(metaclass=ABCMeta):
    """
    Generates text representations of expression objects.

    This is an abstract base class; use :meth:`default` to create a default expression
    formatter.
    """

    @abstractmethod
    def to_text(self, expression: "Expression") -> str:
        """
        Construct a text representation of the given expression.

        :param expression: the expression to represent as text
        :return: a text representation of the expression
        """
        pass

    @staticmethod
    def default(single_line: bool) -> "ExpressionFormatter":
        """
        Get the default expression formatter, formatting expression objects as
        `Python` expressions with correct spacing and formatting.

        :param single_line: if ``True``, return a formatter that does not generate line
            breaks
        :return: a :class:`.PythonExpressionFormatter` with default parameters for all
            formatting settings except for ``single_line``
        """
        from .formatter import PythonExpressionFormatter

        return PythonExpressionFormatter(single_line=single_line)


class HasExpressionRepr(metaclass=ABCMeta):
    """
    Mix-in class for arbitrary classes, rendering :func:`repr` and :class:`str`
    representations using expression objects.
    """

    @abstractmethod
    def to_expression(self) -> "Expression":
        """
        Render this object as an expression.

        :return: the expression representing this object
        """
        pass

    def __repr__(self) -> str:
        # get the expression representing this object, and use the default formatter
        # to get a textual representation of the expression
        return repr(self.to_expression())

    def __str__(self) -> str:
        # get the expression representing this object, and use the default formatter
        # to convert the expression to a formatted string
        return str(self.to_expression())


@inheritdoc(match="[see superclass]")
class Expression(HasExpressionRepr, metaclass=ABCMeta):
    """
    A representation of an expression.
    """

    @property
    @abstractmethod
    def precedence_(self) -> int:
        """
        The precedence of this expression, used to determine the need for parentheses.
        """
        pass

    @property
    @abstractmethod
    def subexpressions_(self) -> Tuple["Expression", ...]:
        """
        The subexpressions of this expression.
        """

    def or_(self, other: "Expression") -> "Expression":
        """
        Create a logical `or` expression using this and another expression as operands.

        :param other: other operand to combine with this expression using a logical `or`
        :return: the logical `or` expression
        """
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.OR, self, other)

    def and_(self, other: "Expression") -> "Expression":
        """
        Create a logical `and` expression using this and another expression as operands.

        :param other: other operand to combine with this expression using a logical
            `and`
        :return: the logical `and` expression
        """
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.AND, self, other)

    def not_(self) -> "Expression":
        """
        Create a logical `not` expression using this expression as the operand.

        :return: the logical `not` expression
        """
        from .composite import UnaryOperation

        return UnaryOperation(UnaryOperator.NOT, self)

    def eq_(self, other: "Expression") -> bool:
        """
        Compare this expression with another for equality.

        For using Python's native equality operator ``==``, see
        :class:`.FrozenExpression`.

        :param other: the expression to compare this expression with
        :return: ``True`` if both expressions are equal; ``False`` otherwise
        """
        self_type = type(self)
        other_type = type(other)

        if self_type is ExpressionAlias:
            self: ExpressionAlias
            return other.eq_(self.expression_)
        elif other_type is ExpressionAlias:
            other: ExpressionAlias
            return self.eq_(other.expression_)
        else:
            # noinspection PyProtectedMember
            return self_type is other_type and self._eq_same_type(other)

    @abstractmethod
    def hash_(self) -> int:
        """
        Calculate the hash for this expression.

        For using Python's native ``hash`` function, see :class:`.FrozenExpression`.
        """
        pass

    def to_expression(self) -> "Expression":
        """[see superclass]"""
        return self

    @abstractmethod
    def _eq_same_type(self: T, other: T) -> bool:
        # assuming other is the same type as self, check if self and other are equal
        pass

    def __add__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.ADD, self, other)

    def __sub__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.SUB, self, other)

    def __mul__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.MUL, self, other)

    def __matmul__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.MATMUL, self, other)

    def __truediv__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.DIV, self, other)

    def __floordiv__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.FLOOR_DIV, self, other)

    def __mod__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.MOD, self, other)

    def __pow__(self, power: Any, modulo=None) -> "Expression":
        if modulo is not None:
            return NotImplemented
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.POW, self, power)

    def __lshift__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.LSHIFT, self, other)

    def __rshift__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.RSHIFT, self, other)

    def __and__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.AND_BITWISE, self, other)

    def __xor__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.XOR_BITWISE, self, other)

    def __or__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.OR_BITWISE, self, other)

    def __radd__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.ADD, other, self)

    def __rsub__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.SUB, other, self)

    def __rmul__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.MUL, other, self)

    def __rmatmul__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.MATMUL, other, self)

    def __rtruediv__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.DIV, other, self)

    def __rfloordiv__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.FLOOR_DIV, other, self)

    def __rmod__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.MOD, other, self)

    def __rpow__(self, power: Any, modulo=None) -> "Expression":
        if modulo is not None:
            return NotImplemented
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.POW, (power, self))

    def __rlshift__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.LSHIFT, other, self)

    def __rrshift__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.RSHIFT, other, self)

    def __rand__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.AND_BITWISE, other, self)

    def __rxor__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.XOR_BITWISE, other, self)

    def __ror__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.OR_BITWISE, other, self)

    def __neg__(self) -> "Expression":
        from .composite import UnaryOperation

        return UnaryOperation(UnaryOperator.NEG, self)

    def __pos__(self) -> "Expression":
        from .composite import UnaryOperation

        return UnaryOperation(UnaryOperator.POS, self)

    def __invert__(self) -> "Expression":
        from .composite import UnaryOperation

        return UnaryOperation(UnaryOperator.INVERT, self)

    def __eq__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.EQ, self, other)

    def __ne__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.NEQ, self, other)

    def __gt__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.GT, self, other)

    def __ge__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.GE, self, other)

    def __lt__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.LT, self, other)

    def __le__(self, other: Any) -> "Expression":
        from .composite import BinaryOperation

        return BinaryOperation(BinaryOperator.LE, self, other)

    def __call__(self, *args: Any, **kwargs: Any) -> "Expression":
        """
        Generate a :class:`.Call` expression.

        :return: the resulting expression
        """
        from .composite import Call

        return Call(self, *args, **kwargs)

    def __getitem__(self, key: Any) -> "Expression":
        from .composite import Index

        return Index(self, key)

    def __setitem__(self, key: Any, value: Any) -> None:
        raise TypeError(f"cannot set indexed item of Expression: {to_list(key)}")

    def __delitem__(self, key: Any) -> None:
        raise TypeError(f"cannot delete indexed item of Expression: {to_list(key)}")

    def __getattr__(self, key: str) -> "Expression":
        if key.startswith("_") or key.endswith("_"):
            raise AttributeError(key)
        else:
            from .composite import Attr

            return Attr(obj=self, attribute=key)

    def __setattr__(self, key: str, value: Any) -> None:
        if key.startswith("_") or key.endswith("_"):
            super().__setattr__(key, value)
        else:
            raise TypeError(f"cannot set public field of Expression: {key}")

    def __iter__(self) -> None:
        # we need to rule iteration out explicitly, otherwise we'd get infinite 'for'
        # loops through iterating via __getitem__
        raise TypeError(f"'{type(self).__name__}' object is not iterable: {repr(self)}")

    def __repr__(self) -> str:
        # get a textual representation of the expression using the default formatter
        return ExpressionFormatter.default(single_line=True).to_text(self)

    def __str__(self) -> str:
        # convert the expression to a formatted string using the default formatter
        return ExpressionFormatter.default(single_line=False).to_text(self)

    __hash__ = None


class FrozenExpression(HasExpressionRepr):
    """
    A frozen expression.

    Frozen expressions cannot be used as subexpressions in other expressions,
    but instead support Python's native semantics for equality and hashing.
    """

    def __init__(self, expression: Expression) -> None:
        """
        :param expression: the expression to be frozen
        """
        self._expression = expression

    def to_expression(self) -> "Expression":
        """
        Get the underlying un-frozen expression.

        :return: the underlying un-frozen expression
        """
        return self._expression

    def __eq__(self, other: "FrozenExpression") -> bool:
        return type(self) is type(other) and self._expression.eq_(other._expression)

    def __hash__(self) -> int:
        return self._expression.hash_()


def make_expression(value: Any) -> Expression:
    """
    Convert a Python :class:`object` into an expression.

    Expressions are generated as follows:

    - Instances of :class:`HasExpressionRepr` are converted by calling method
      :meth:`~.HasExpressionRepr.to_expression`
    - Python literals (strings, numbers) are converted to :class:`Lit` expressions
    - Python tuples, lists, sets, and dictionaries are converted to equivalent
      expressions, and all elements are converted to expressions recursively
    - Other iterables are converted to a :class:`Call` expression, with the
      name of the iterable as the callee, and its elements as the arguments
    - Other objects implementing the ``__name__`` attribute are converted to an
      :class:`Id` expression with the same name
    - All other values are converted to a :class:`Lit` expression

    :param value: value to convert to an expression
    :return: the resulting expression
    """

    if isinstance(value, HasExpressionRepr):
        return value.to_expression()
    elif isinstance(value, str):
        from .atomic import Lit

        return Lit(value)
    elif isinstance(value, list):
        from .composite import ListLiteral

        return ListLiteral(*value)
    elif isinstance(value, tuple):
        from .composite import TupleLiteral

        return TupleLiteral(*value)
    elif isinstance(value, set):
        from .composite import SetLiteral

        return SetLiteral(*value)
    elif isinstance(value, dict):
        from .composite import DictLiteral

        return DictLiteral(*value.items())
    elif isinstance(value, slice):
        from .atomic import Epsilon
        from .composite import BinaryOperation

        args = [
            Epsilon() if value is None else value
            for value in (value.start, value.stop, value.step)
        ]
        if value.step is not None:
            return BinaryOperation(BinaryOperator.SLICE, *args)
        else:
            return BinaryOperation(BinaryOperator.SLICE, args[0], args[1])
    elif isinstance(value, Iterable):
        from .atomic import Id
        from .composite import Call

        return Call(Id(type(value)), *value)
    else:
        name: Optional[str] = getattr(value, "__name__", None)
        if name:
            from .atomic import Id

            return Id(value)
        else:
            from .atomic import Lit

            return Lit(value)


def freeze(expression: Expression) -> FrozenExpression:
    """
    Convenience function to freeze an expression.

    Equivalent to ``FrozenExpression(expression)``.

    :param expression: the expression to freeze
    :return: the resulting frozen expression
    """
    return FrozenExpression(expression)


#
# Expression alias
#


@inheritdoc(match="[see superclass]")
class ExpressionAlias(Expression):
    """
    An alias pointing to another expression, and representing that expression.

    Useful for substituting subexpressions of expressions without having to reconstruct
    the surrounding expression.

    When comparing or hashing expressions, an expression alias transparently defers to
    the expression it represents, hence expressions containing aliases are
    indistinguishable from their non-aliased counterparts.
    """

    def __init__(self, expression: Expression) -> None:
        """
        :param expression: the expression referred to by this alias
        """
        super().__init__()
        self._expression = expression

    @property
    def expression_(self) -> Expression:
        """
        The expression represented by this alias; may be set to a new expression.
        """
        return self._expression

    @expression_.setter
    def expression_(self, expression: Expression) -> None:
        """
        Set the expression represented by this alias.

        :param expression: the expression to be set
        """
        self._expression = expression

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return self.expression_.precedence_

    @property
    def subexpressions_(self) -> Tuple["Expression", ...]:
        """[see superclass]"""
        return (self._expression,)

    def hash_(self) -> int:
        """[see superclass]"""
        return self.expression_.hash_()

    def _eq_same_type(self, other: "ExpressionAlias") -> bool:
        return self._expression.eq_(other._expression)


__tracker.validate()



"""
Dendrogram styles.
"""

import logging
from typing import Optional, Sequence, TextIO

from .. import TextStyle
from ..color import text_contrast_color
from .base import DendrogramMatplotStyle, DendrogramStyle
from pytools.api import AllTracker, inheritdoc
from pytools.text import CharacterMatrix

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = [
    "DendrogramLineStyle",
    "DendrogramHeatmapStyle",
    "DendrogramReportStyle",
]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


@inheritdoc(match="[see superclass]")
class DendrogramLineStyle(DendrogramMatplotStyle):
    """
    Draws dendrograms as "classical" trees, using coloring lines along a color gradient
    to indicate leaf/branch weights.
    """

    @classmethod
    def get_default_style_name(cls) -> str:
        """[see superclass]"""
        return f"{super().get_default_style_name()}_line"

    def draw_link_leg(
        self, bottom: float, top: float, leaf: float, weight: float, tree_height
    ) -> None:
        """[see superclass]"""
        self._draw_line(x1=bottom, x2=top, y1=leaf, y2=leaf, weight=weight)

    def draw_link_connector(
        self,
        bottom: float,
        top: float,
        first_leaf: int,
        n_leaves_left: int,
        n_leaves_right: int,
        weight: float,
        tree_height: float,
    ) -> None:
        """[see superclass]"""
        self._draw_line(
            x1=bottom,
            x2=bottom,
            y1=first_leaf + (n_leaves_left - 1) / 2,
            y2=first_leaf + n_leaves_left + (n_leaves_right - 1) / 2,
            weight=weight,
        )

        self.draw_link_leg(
            bottom=bottom,
            top=top,
            leaf=(first_leaf + (n_leaves_left + n_leaves_right - 1) / 2),
            weight=weight,
            tree_height=0,
        )

    def _draw_line(
        self, x1: float, x2: float, y1: float, y2: float, weight: float
    ) -> None:
        self.ax.plot((x1, x2), (y1, y2), color=self.color_for_value(weight))


@inheritdoc(match="[see superclass]")
class DendrogramHeatmapStyle(DendrogramMatplotStyle):
    """
    Draws dendrograms as a tree composed of color tiles, coloring tiles as a heat map
    indicating leaf/branch weights.
    """

    def draw_link_leg(
        self, bottom: float, top: float, leaf: int, weight: float, tree_height: float
    ) -> None:
        """[see superclass]"""
        self._draw_hbar(x=bottom, w=top - bottom, y=leaf, h=1, weight=weight)

    def draw_link_connector(
        self,
        bottom: float,
        top: float,
        first_leaf: int,
        n_leaves_left: int,
        n_leaves_right: int,
        weight: float,
        tree_height: float,
    ) -> None:
        """[see superclass]"""
        self._draw_hbar(
            x=bottom,
            w=top - bottom,
            y=first_leaf,
            h=n_leaves_left + n_leaves_right,
            weight=weight,
        )

    def start_drawing(self, title: str, **kwargs) -> None:
        """[see superclass]"""
        super().start_drawing(title=title, **kwargs)
        self.ax.margins(0, 0)

    def _draw_hbar(self, x: float, y: float, w: float, h: float, weight: float) -> None:
        fill_color = self.color_for_value(weight)

        self.ax.barh(
            y=[y - 0.5],
            width=[w],
            height=[h],
            left=[x],
            align="edge",
            color=fill_color,
            edgecolor=self.colors.foreground,
            linewidth=1,
        )

        weight_percent = weight * 100
        label = (
            f"{weight_percent:.2g}%"
            if weight_percent < 99.5
            else f"{round(weight_percent):.3g}%"
        )

        x_text = x + w / 2
        y_text = y + (h - 1) / 2
        text_width, _ = self.text_dimensions(text=label, x=x_text, y=y_text)

        if text_width <= w:
            self.ax.text(
                x_text,
                y_text,
                label,
                ha="center",
                va="center",
                color=text_contrast_color(fill_color),
            )


@inheritdoc(match="[see superclass]")
class DendrogramReportStyle(TextStyle, DendrogramStyle):
    """
    Renders dendrograms as ASCII graphics for inclusion in plain-text reports.
    """

    #: The default width of labels.
    DEFAULT_LABEL_WIDTH = 20

    #: The number of characters that will be allocated for the label column,
    #: including the weight.
    label_width: int

    #: Maximum number of text lines to output including the title;
    #: additional lines of the dendrogram will be clipped (optional).
    max_height: int

    def __init__(
        self,
        out: TextIO = None,
        width: int = 80,
        label_width: Optional[int] = None,
        max_height: int = 100,
    ) -> None:
        """
        :param label_width: the number of characters that will be allocated for the
            label column, including the weight (optional; defaults to
            %DEFAULT_LABEL_WIDTH% characters or half of arg `width`,
            whichever is smaller)
        :param max_height: maximum number of text lines to output including the title;
            additional lines of the dendrogram will be clipped (default: 100)
        """
        super().__init__(out=out, width=width)
        if max_height <= 0:
            raise ValueError(
                f"arg max_height={max_height} expected to be a positive integer"
                f" {max_height}"
            )
        if label_width is not None and label_width > width // 2:
            raise ValueError(
                f"arg label_width={label_width} must be half or less of arg "
                f"width={width}"
            )
        self.max_height = max_height
        self.label_width = (
            min(DendrogramReportStyle.DEFAULT_LABEL_WIDTH, width // 2)
            if label_width is None
            else label_width
        )
        self._dendrogram_right = width - self.label_width
        self._char_matrix = None
        self._n_labels = None

    __init__.__doc__ = TextStyle.__init__.__doc__ + __init__.__doc__.replace(
        "%DEFAULT_LABEL_WIDTH%", str(DEFAULT_LABEL_WIDTH)
    )

    def draw_leaf_names(self, *, names: Sequence[str]) -> None:
        """[see superclass]"""

        matrix = self._char_matrix
        n_labels = len(names)
        if n_labels > self.max_height:
            n_labels = self.max_height - 1
            matrix[n_labels, :] = f"{'clipped':~^{self.width}s}\n"
        self._n_labels = n_labels
        label_width = self.__weight_column
        for row, label in enumerate(names[:n_labels]):
            matrix[row, :label_width] = label + " "

    def draw_link_leg(
        self, bottom: float, top: float, leaf: float, weight: float, tree_height: float
    ) -> None:
        """[see superclass]"""

        # determine the y coordinate in the character matrix
        line_y = int(leaf)

        # if leaf is in between two leaves, we want to draw a line in
        # between two text lines (using an underscore symbol)
        is_in_between_line = round(leaf * 2) & 1

        # draw the link leg in the character matrix
        self._char_matrix[
            line_y + is_in_between_line,
            self._x_pos(bottom, tree_height) : self._x_pos(top, tree_height),
        ] = (
            "_" if is_in_between_line else "-"
        )

        # if we're in a leaf, we can draw the weight next to he label
        if bottom == 0:
            self._char_matrix[
                line_y, self.__weight_column : self.label_width
            ] = f"{weight * 100:3.0f}%"

    def draw_link_connector(
        self,
        bottom: float,
        top: float,
        first_leaf: int,
        n_leaves_left: int,
        n_leaves_right: int,
        weight: float,
        tree_height: float,
    ) -> None:
        """[see superclass]"""

        y1 = first_leaf + n_leaves_left // 2
        y2 = first_leaf + n_leaves_left + (n_leaves_right - 1) // 2

        self.draw_link_leg(
            bottom=bottom,
            top=top,
            leaf=(first_leaf - 0.5) + (n_leaves_left + n_leaves_right) / 2,
            weight=weight,
            tree_height=tree_height,
        )

        x = self._x_pos(bottom, tree_height)
        matrix = self._char_matrix
        if y2 - y1 > 1:
            matrix[(y1 + 1) : y2, x] = "|"
        matrix[y1, x] = "/"
        matrix[y2, x] = "\\"

    def start_drawing(self, title: str, **kwargs) -> None:
        """[see superclass]"""
        super().start_drawing(title=title, **kwargs)
        self._char_matrix = CharacterMatrix(
            n_rows=self.max_height, n_columns=self.width
        )

    def finalize_drawing(self, **kwargs) -> None:
        """[see superclass]"""
        try:
            super().finalize_drawing(**kwargs)
            for row in reversed(range(self._n_labels + 1)):
                self.out.write(f"{self._char_matrix[row, :]}\n")
        finally:
            self._char_matrix = None
            self._n_labels = None

    def _x_pos(self, h: float, h_max: float) -> int:
        # calculate the horizontal position in the character grid,
        # ensuring that h=h_max still yields a position inside the grid (factor 0.99999)
        return self.label_width + int(self._dendrogram_right * h / h_max * 0.99999)

    @property
    def __weight_column(self) -> int:
        return self.label_width - 5


__tracker.validate()



"""
Composite expressions.

A composite expressions combines subexpressions to form a new, combines expression.

Composite expressions comprise

- operations: e.g., ``-<expression>``, ``<expression 1> - <expression 2>``
- tuple, list, set, dictionary literals
- call expressions: ``<expression 1>(<expression 2>])``
- indexing expressions: ``<expression 1>[<expression 2>]``
- attribute access: ``<expression 1>.<attribute name>``
- lambda expressions: ``lambda [x, y, …]: <expression>``
"""

from ._composite import *



"""
Core implementation of :mod:`pytools.viz.distribution`.
"""

import logging
from typing import Iterable, Optional, Sequence, Type, Union

import numpy as np
import pandas as pd

from .. import Drawer, MatplotStyle
from .base import ECDF, ECDFStyle, XYSeries
from pytools.api import AllTracker, inheritdoc

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = [
    "ECDFMatplotStyle",
    "ECDFDrawer",
]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class ECDFMatplotStyle(ECDFStyle, MatplotStyle):
    """
    Draws ECDF plots using `matplotlib`.
    """

    def _draw_ecdf(
        self, ecdf: ECDF, x_label: str, iqr_multiple: float, iqr_multiple_far: float
    ) -> None:
        def _iqr_annotation(multiple: float) -> str:
            return f"(> {multiple:.3g} * IQR)"

        ax = self.ax
        matplotlib_kwargs = {"marker": ".", "linestyle": "none"}
        ax.plot(
            ecdf.inliers.x,
            ecdf.inliers.y,
            color=self.colors.status_ok,
            label="inlier",
            **matplotlib_kwargs,
        )
        ax.plot(
            ecdf.outliers.x,
            ecdf.outliers.y,
            color=self.colors.status_warning,
            label=f"outlier {_iqr_annotation(multiple=iqr_multiple)}",
            **matplotlib_kwargs,
        )
        ax.plot(
            ecdf.far_outliers.x,
            ecdf.far_outliers.y,
            color=self.colors.status_critical,
            label=f"far outlier {_iqr_annotation(multiple=iqr_multiple_far)}",
            **matplotlib_kwargs,
        )

        # add axis labels and legend
        self.ax.set_xlabel(x_label)
        ax.set_ylabel("count")
        ax.legend()


@inheritdoc(match="[see superclass]")
class ECDFDrawer(Drawer[Sequence[float], ECDFStyle]):
    """
    Drawer for empirical cumulative density functions (ECDFs), highlighting
    outliers using Tukey's outlier test.

    The drawer highlights samples as `outliers` or `far outliers`.

    A sample is considered an outlier if it is outside the range
    :math:`[q_1 - m * \\mathit{iqr}, q_3 + m * \\mathit{iqr}]`
    where :math:`q_1` and :math:`q_3` are the lower and upper quartiles,
    :math:`\\mathit{iqr} = q3 - q1` is the `inter-quartile range (IQR)`, and
    :math:`m` is the `IQR multiple`.

    By convention, common values for :math:`m` are :math:`m = 1.5` for outliers,
    and :math:`m = 3` for far outliers.
    """

    #: iqr multiple to determine outliers;
    #: if ``None``, then no outliers and far outliers are computed
    iqr_multiple: Optional[float]

    #: iqr multiple to determine far outliers;
    #: if ``None``, then no far outliers are computed, otherwise
    #: must be greater than ``iqr_multiple``
    iqr_multiple_far: Optional[float]

    #: if ``True``, do not plot far outliers
    hide_far_outliers: bool

    def __init__(
        self,
        style: Optional[Union[ECDFStyle, str]] = None,
        iqr_multiple: Optional[float] = 1.5,
        iqr_multiple_far: Optional[float] = 3.0,
        hide_far_outliers: bool = False,
    ) -> None:
        """
        :param iqr_multiple: iqr multiple to determine outliers; if ``None``, then no
            outliers and far outliers are computed.
        :param iqr_multiple_far: iqr multiple to determine far outliers; if ``None``,
            then no far outliers are computed, otherwise must be greater than
            `iqr_multiple`.
        :param hide_far_outliers: if ``True``, do not plot far outliers
            (default: ``False``)
        """
        super().__init__(style=style)

        if iqr_multiple_far:
            if not iqr_multiple:
                raise ValueError(
                    "arg iqr_multiple must be defined if arg iqr_multiple_far is "
                    "defined"
                )
            if iqr_multiple_far <= iqr_multiple:
                raise ValueError(
                    f"arg iqr_multiple={iqr_multiple} must be smaller than "
                    f"arg iqr_multiple_far={iqr_multiple_far}"
                )

        self.iqr_multiple = iqr_multiple
        self.iqr_multiple_far = iqr_multiple_far
        self.hide_far_outliers = hide_far_outliers

    def draw(self, data: Sequence[float], title: Optional[str] = None) -> None:
        """
        Draw the ECDF.

        :param data: the data to draw
        :param title: the title of the chart (optional; defaults to ``"ECDF"``; if arg
            `data` is a :class:`~.Series` or any other class with a `name`
            attribute, then the default title will include the name)
        """
        if title is None:
            if hasattr(data, "name"):
                title = f"ECDF: {data.name}"
            else:
                title = "ECDF"
        super().draw(data=data, title=title)

    @classmethod
    def get_style_classes(cls) -> Iterable[Type[ECDFStyle]]:
        """[see superclass]"""
        return [
            ECDFMatplotStyle,
        ]

    def _draw(self, data: Sequence[float]) -> None:
        ecdf = self._ecdf(data=data)
        x_label = getattr(data, "name", "value")
        # noinspection PyProtectedMember
        self.style._draw_ecdf(
            ecdf=ecdf,
            x_label=x_label,
            iqr_multiple=self.iqr_multiple,
            iqr_multiple_far=self.iqr_multiple_far,
        )

    def _ecdf(self, data: Sequence[float]) -> ECDF:
        """
        Compute ECDF for scalar values.

        Return the x and y values of an empirical cumulative distribution plot of the
        values in arg `data`. Outlier and far outlier points are returned in separate
        lists.

        :param data: the series of values forming our sample
        :return: x_inlier, y_inlier, x_outlier, y_outlier, x_far_outlier,
            y_far_outlier:
            lists of x and y coordinates for the ecdf plot for the inlier, outlier and
            far outlier points.
        """

        # x-data for the ECDF: x
        if isinstance(data, pd.Series):
            data = data.values
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        x = np.sort(data[~np.isnan(data)])

        # Number of data points: n
        n = len(x)

        # y-data for the ECDF: y
        y = np.arange(1, n + 1)

        iqr_multiple = self.iqr_multiple
        iqr_multiple_far = self.iqr_multiple_far
        if iqr_multiple:
            # outliers
            q1, q3 = np.percentile(x, [25, 75])

            iqr = q3 - q1
            out_lower = q1 - iqr_multiple * iqr
            out_upper = q3 + iqr_multiple * iqr
            inlier_mask = (x >= out_lower) & (x <= out_upper)

            if iqr_multiple_far:
                far_out_lower = q1 - iqr_multiple_far * iqr
                far_out_upper = q3 + iqr_multiple_far * iqr
                outlier_mask = (
                    (~inlier_mask) & (x >= far_out_lower) & (x <= far_out_upper)
                )
                far_out_mask = ~(inlier_mask | outlier_mask)
            else:
                outlier_mask = ~inlier_mask
                far_out_mask = []

        else:
            inlier_mask = slice(None)
            outlier_mask = []
            far_out_mask = []

        return ECDF(
            inliers=XYSeries(x=x[inlier_mask], y=y[inlier_mask]),
            outliers=XYSeries(x=x[outlier_mask], y=y[outlier_mask]),
            far_outliers=(
                XYSeries(x=[], y=[])
                if self.hide_far_outliers
                else XYSeries(x=x[far_out_mask], y=y[far_out_mask])
            ),
        )


__tracker.validate()



"""
Implementation of :mod:`pytools.expression.composite.base`.
"""
import logging
from typing import Any

from ... import Expression
from ...atomic import Epsilon, Id
from ...base import SimplePrefixExpression
from ...operator import BinaryOperator, UnaryOperator
from pytools.api import AllTracker, inheritdoc

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = [
    "KeywordArgument",
    "DictEntry",
    "LambdaDefinition",
]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Class definitions
#


@inheritdoc(match="[see superclass]")
class KeywordArgument(SimplePrefixExpression):
    """
    A keyword argument, used by functions.
    """

    _PRECEDENCE = BinaryOperator.EQ.precedence

    def __init__(self, name: str, value: Any):
        """
        :param name: the name of the keyword
        :param value: the value for the keyword
        """
        super().__init__(prefix=Id(name), body=value)
        self._name = name

    @property
    def name_(self) -> str:
        """
        The name of this keyword argument.
        """
        return self._name

    @property
    def separator_(self) -> str:
        """[see superclass]"""
        return "="

    @property
    def value_(self) -> Expression:
        """
        The name of this keyword argument.
        """
        return self.body_

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return self._PRECEDENCE


@inheritdoc(match="[see superclass]")
class DictEntry(SimplePrefixExpression):
    """
    Two expressions separated by a colon, used in dictionaries.
    """

    _PRECEDENCE = BinaryOperator.COLON.precedence

    def __init__(self, key: Any, value: Any):
        """
        :param key: the key of the dictionary entry
        :param value: the value of the dictionary entry
        """
        super().__init__(prefix=key, body=value)

    @property
    def key_(self) -> Expression:
        """
        The key of this dictionary entry; identical with the expression prefix.
        """
        return self.prefix_

    @property
    def separator_(self) -> str:
        """A ``:``, followed by a space."""
        return ": "

    @property
    def value_(self) -> Expression:
        """
        The value of this dictionary entry; identical with the expression body.
        """
        return self.body_

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return DictEntry._PRECEDENCE


@inheritdoc(match="[see superclass]")
class LambdaDefinition(SimplePrefixExpression):
    """
    Function parameters and body separated by a colon, used inside lambda expressions.
    """

    _PRECEDENCE = UnaryOperator.LAMBDA.precedence

    def __init__(self, *params: Id, body: Any):
        """
        :param params: the parameters of the lambda expression
        :param body: the body of the lambda expression
        """
        if not params:
            params_expression = Epsilon()
        elif len(params) == 1:
            params_expression = params[0]
        else:
            from .. import BinaryOperation

            params_expression = BinaryOperation(BinaryOperator.COMMA, *params)
        super().__init__(prefix=params_expression, body=body)

    @property
    def params_(self) -> Expression:
        """
        The parameters of the lambda expression.
        """
        return self.prefix_

    @property
    def separator_(self) -> str:
        """A ``:``, followed by a space."""
        return ": "

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return LambdaDefinition._PRECEDENCE



"""
Implementation of sphinx module.
"""
import importlib
import itertools
import logging
import re
from abc import ABCMeta, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    TypeVar,
    cast,
)

import typing_inspect

from pytools.api import AllTracker, get_generic_bases, inheritdoc, public_module_prefix

try:
    # re.Pattern was introduced in Python 3.7 …
    Pattern = re.Pattern
except AttributeError:
    # … in older versions, we set it to the wildcard type
    Pattern = Any

try:
    # ForwardRef was introduced in Python 3.7 …
    # noinspection PyUnresolvedReferences
    from typing import ForwardRef
except ImportError:
    # … in Python 3.6, it is called _ForwardRef
    # noinspection PyProtectedMember,PyUnresolvedReferences
    from typing import _ForwardRef as ForwardRef

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = [
    "SphinxCallback",
    "AutodocProcessDocstring",
    "AutodocBeforeProcessSignature",
    "AutodocProcessSignature",
    "AutodocSkipMember",
    "AddInheritance",
    "CollapseModulePaths",
    "CollapseModulePathsInDocstring",
    "CollapseModulePathsInSignature",
    "SkipIndirectImports",
    "Replace3rdPartyDoc",
]

#
# Type variables
#

#: Mock type declaration: Sphinx application object.
Sphinx = Any


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class SphinxCallback(metaclass=ABCMeta):
    """
    Abstract base class for autodoc processors.

    Can be passed to :meth:`~sphinx.application.Sphinx.connect` as a callback for an
    event.
    """

    @property
    @abstractmethod
    def event(self) -> str:
        """
        The event processed by this callback.

        Used to connect this callback to sphinx in method :meth:`.connect`.

        :return: name of the event
        """
        pass

    def connect(self, app: Sphinx, priority: Optional[int] = None) -> int:
        """
        Register this callback to be called when :attr:`.event` is emitted.

        Registered callbacks will be invoked on event in the order of priority and
        registration. The priority is ascending order.

        :param app: the Sphinx application to register this processor with
        :param priority: the priority of this processor
        :return: a listener ID that can be used as an argument to
            :meth:`~sphinx.application.Sphinx.disconnect`.
        """
        if priority is None:
            return app.connect(event=self.event, callback=self)
        else:
            return app.connect(event=self.event, callback=self, priority=priority)


class AutodocProcessDocstring(SphinxCallback, metaclass=ABCMeta):
    """
    An autodoc processor for docstrings.
    """

    @property
    def event(self) -> str:
        """
        ``"autodoc-process-docstring"``
        """
        return "autodoc-process-docstring"

    @abstractmethod
    def process(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        options: object,
        lines: List[str],
    ) -> None:
        """
        Process an event.

        :param app: the Sphinx application object
        :param what: the type of the object which the docstring belongs to (one of
            "module", "class", "exception", "function", "method", "attribute")
        :param name: the fully qualified name of the object
        :param obj: the object itself
        :param options: the options given to the directive: an object with attributes
            ``inherited_members``, ``undoc_members``, ``show_inheritance`` and
            ``noindex`` that are ``True`` if the flag option of same name was given to
            the auto directive
        :param lines: the lines of the docstring
        """
        pass

    def __call__(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        options: object,
        lines: List[str],
    ) -> None:
        try:
            self.process(
                app=app, what=what, name=name, obj=obj, options=options, lines=lines
            )
        except Exception as e:
            log.error(e)
            raise


class AutodocBeforeProcessSignature(SphinxCallback, metaclass=ABCMeta):
    """
    An autodoc processor invoked before processing signatures.
    """

    @property
    def event(self) -> str:
        """
        ``"autodoc-before-process-signature"``
        """
        return "autodoc-before-process-signature"

    @abstractmethod
    def process(self, app: Sphinx, obj: object, bound_method: bool) -> None:
        """
        Process an event.

        :param app: the Sphinx application object
        :param obj: the object itself
        :param bound_method: a boolean indicates an object is bound method or not
        """
        pass

    def __call__(
        self, app: Sphinx, obj: object, bound_method: bool
    ) -> Optional[Tuple[str, str]]:
        try:
            return self.process(app=app, obj=obj, bound_method=bound_method)
        except Exception as e:
            log.error(e)


class AutodocProcessSignature(SphinxCallback, metaclass=ABCMeta):
    """
    An autodoc processor for processing signatures.
    """

    @property
    def event(self) -> str:
        """
        ``"autodoc-process-signature"``
        """
        return "autodoc-process-signature"

    @abstractmethod
    def process(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        options: object,
        signature: Optional[str],
        return_annotation: Optional[str],
    ) -> Optional[Tuple[str, str]]:
        """
        Process an event.

        :param app: the Sphinx application object
        :param what: the type of the object which the docstring belongs to (one of
            "module", "class", "exception", "function", "method", "attribute")
        :param name: the fully qualified name of the object
        :param obj: the object itself
        :param options: the options given to the directive: an object with attributes
            ``inherited_members``, ``undoc_members``, ``show_inheritance`` and
            ``noindex`` that are ``True`` if the flag option of same name was given to
            the auto directive
        :param signature: function signature, as a string of the form
            ``(parameter_1, parameter_2)``, or ``None`` if introspection did not succeed
            and signature was not specified in the directive
        :param return_annotation: function return annotation as a string of the form
            `` -> <annotation>``, or ``None`` if there is no return annotation
        """
        pass

    def __call__(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        options: object,
        signature: Optional[str],
        return_annotation: Optional[str],
    ) -> Optional[Tuple[str, str]]:
        try:
            return self.process(
                app=app,
                what=what,
                name=name,
                obj=obj,
                options=options,
                signature=signature,
                return_annotation=return_annotation,
            )
        except Exception as e:
            log.error(e)


class AutodocSkipMember(SphinxCallback, metaclass=ABCMeta):
    """
    An autodoc-skip-member processor.
    """

    @property
    def event(self) -> str:
        """
        ``"autodoc-skip-member"``
        """
        return "autodoc-skip-member"

    @abstractmethod
    def process(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        skip: bool,
        options: object,
    ) -> Optional[bool]:
        """
        Decide whether a member should be included in the documentation.

        :param app: the Sphinx application object
        :param what: the type of the object which the docstring belongs to (one of
            "module", "class", "exception", "function", "method", "attribute")
        :param name: the fully qualified name of the object
        :param obj: the object itself
        :param skip: a boolean indicating if autodoc will skip this member if the user
            handler does not override the decision
        :param options: the options given to the directive: an object with attributes
            ``inherited_members``, ``undoc_members``, ``show_inheritance`` and
            ``noindex`` that are ``True`` if the flag option of same name was given to
            the auto directive
        :return: ``True`` if the member should be excluded; ``False`` if the member
            should be included; ``None`` to fall back to the skipping behavior of
            autodoc and other enabled extensions


        """
        pass

    def __call__(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        skip: bool,
        options: object,
    ) -> Optional[bool]:
        try:
            return self.process(
                app=app, what=what, name=name, obj=obj, skip=skip, options=options
            )
        except Exception as e:
            log.error(e)
            raise


@inheritdoc(match="[see superclass]")
class AddInheritance(AutodocProcessDocstring):
    """
    Add list of base classes as the first line of the docstring.

    Ignore builtin classes and classes that have already been visited once before.
    """

    def __init__(self, collapsible_submodules: Mapping[str, str]):
        """
        :param collapsible_submodules: mapping of submodule paths to shorter
            *(collapsed)* versions they should be replaced with
        """
        super().__init__()
        self.collapsible_submodules = collapsible_submodules

        #: Dict mapping visited classes to their unprocessed docstrings.
        self._visited: Dict[type, str] = {}

    F_BASES = ":bases:"  #: Field directive for base classes.
    F_GENERICS = ":generic types:"  #: Field directive for generic types.
    F_METACLASSES = ":metaclasses:"  #: Field directive for metaclasses.

    def process(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        options: object,
        lines: List[str],
    ) -> None:
        """[see superclass]"""

        if what == "class":

            # add bases and generics documentation to class

            # generate the RST for bases and generics
            class_ = cast(type, obj)

            _current_lines = "\n".join(lines)
            try:
                _seen_lines = self._visited[class_]
                if _current_lines != _seen_lines:
                    # we are seeing another part of the docstring, probably in __init__
                    # vs. the class docstring;
                    # ignore this to prevent adding the same content at two places
                    return
            except KeyError:
                # we are seeing a class for the first time; store its content so we can
                # detect and allow repeat visits
                self._visited[class_] = _current_lines

            bases_with_origin = [
                (base, typing_inspect.get_origin(base) or base)
                for base in set(self._get_bases(class_, include_subclass=False))
            ]
            bases = [
                base
                for base, origin in bases_with_origin
                if not any(
                    origin is not other and issubclass(other, origin)
                    for _, other in bases_with_origin
                )
            ]

            bases_lines: List[str] = [""]
            if bases:
                base_names = (self._class_name_with_generics(base) for base in bases)
                bases_lines.append(f'{AddInheritance.F_BASES} {", ".join(base_names)}')

            generics: List[str] = self._get_generics(class_)
            if generics:
                bases_lines.append(f'{AddInheritance.F_GENERICS} {", ".join(generics)}')

            metaclasses: List[str] = self._get_metaclasses(class_)
            if metaclasses:
                bases_lines.append(
                    f'{AddInheritance.F_METACLASSES} {", ".join(metaclasses)}'
                )

            bases_lines.append("")

            # insert this after the intro text, and before class parameters

            def _insert_position() -> int:
                for n, line in enumerate(lines):
                    if re.match(r"\s*:\w+(?:\s+\w+)*:", line) and (
                        n == 0 or not line[n - 1].strip()
                    ):
                        return n
                return len(lines)

            if len(bases_lines) > 0:
                pos = _insert_position()
                lines[pos:pos] = bases_lines

    @staticmethod
    def _class_attr(cls: type, attr: str, default: Callable[[], str]) -> str:
        def _get_attr(_cls: type) -> str:
            try:
                # we try to get the class attribute
                return getattr(_cls, attr)
            except AttributeError:
                # if the attribute is not defined, this class is likely to have generic
                # arguments, so we re-try recursively with the origin (unless the origin
                # is the class itself to avoid infinite recursion)
                cls_origin = typing_inspect.get_origin(cls)
                if cls_origin != _cls:
                    return _get_attr(_cls=cls_origin)
                else:
                    # as a last resort, we create the default value
                    return default()

        return _get_attr(_cls=cls)

    def _class_module(self, cls: type) -> str:
        module_name = AddInheritance._class_attr(
            cls=cls,
            attr="__publicmodule__",
            default=lambda: AddInheritance._class_attr(
                cls=cls, attr="__module__", default=lambda: ""
            ),
        )

        # return the collapsed submodule if it exists,
        # else return the unchanged module name
        return self.collapsible_submodules.get(module_name, module_name)

    @staticmethod
    def _class_name(cls: type) -> str:
        return AddInheritance._class_attr(
            cls=cls, attr="__qualname__", default=lambda: str(cls)
        )

    def _full_name(self, cls: type) -> str:
        # get the full name of the class, including the module prefix
        return f"{self._class_module(cls=cls)}.{self._class_name(cls=cls)}"

    def _class_name_with_generics(self, cls: Any) -> str:
        def _class_tag(_name: Any) -> str:
            return f":class:`{str(_name)}`"

        if isinstance(cls, TypeVar):
            return _class_tag(cls)

        if isinstance(cls, ForwardRef):
            if cls.__forward_evaluated__:
                cls = cls.__forward_value__
            else:
                return _class_tag(f".{cls.__forward_arg__}")

        if not hasattr(cls, "__module__"):
            return _class_tag(cls)

        if cls.__module__ in ("__builtin__", "builtins"):
            return _class_tag(cls.__name__)

        else:
            generic_args = [
                self._class_name_with_generics(arg)
                for arg in typing_inspect.get_args(cls, evaluate=True)
            ]

            generic_arg_str = f' [{", ".join(generic_args)}]' if generic_args else ""

            return f":class:`~{self._full_name(cls)}`{generic_arg_str}"

    def _typevar_name(self, cls: TypeVar) -> str:
        if isinstance(cls, TypeVar):
            args: List[str] = [
                self._class_name_with_generics(c) for c in cls.__constraints__
            ]
            if cls.__bound__:
                args.append(f"bound= {self._class_name_with_generics(cls.__bound__)}")
            if cls.__covariant__:
                args.append("*__covariant__=True*")
            if cls.__contravariant__:
                args.append("*__contravariant__=True*")
            return f'{cls}({", ".join(args)})' if args else str(cls)
        else:
            return str(cls)

    def _get_bases(
        self, subclass: type, include_subclass: bool
    ) -> Generator[type, None, None]:
        # get the names of the immediate base classes of arg _subclass

        visited_classes: Set[type] = set()

        def _inner(
            _subclass: type, _include_subclass: bool
        ) -> Generator[type, None, None]:
            # ensure we have the non-generic origin class
            _subclass: type = typing_inspect.get_origin(_subclass) or _subclass

            if _subclass in visited_classes:
                return
            visited_classes.add(_subclass)

            # get the base classes; try generic bases first then fall back to regular
            # bases
            base_classes: Tuple[type] = (
                get_generic_bases(_subclass) or _subclass.__bases__
            )

            # include the _subclass itself in the list of bases, if requested
            if _include_subclass:
                # noinspection PyTypeChecker
                base_classes = (_subclass, *base_classes)

            # get the names of all base classes; go up the class hierarchy in case of
            # hidden classes
            for base in base_classes:

                # exclude object and Generic types
                if base is object or typing_inspect.get_origin(base) is Generic:
                    continue

                # exclude protected classes
                elif self._class_name(base).startswith("_"):
                    yield from _inner(base, _include_subclass=False)

                # all other classes will be listed as bases
                else:
                    yield base

        return _inner(subclass, _include_subclass=include_subclass)

    def _get_generics(self, child_class: type) -> List[str]:
        return list(
            itertools.chain.from_iterable(
                (
                    self._typevar_name(arg)
                    for arg in typing_inspect.get_args(base, evaluate=True)
                )
                for base in get_generic_bases(child_class)
                if typing_inspect.get_origin(base) is Generic
            )
        )

    def _get_metaclasses(self, class_: type) -> List[str]:
        return [
            self._class_name_with_generics(meta_)
            for meta_ in self._get_bases(type(class_), include_subclass=True)
            if meta_ is not type
        ]


class CollapseModulePaths(metaclass=ABCMeta):
    """
    Replace private module paths with their public prefix so that object references
    can be matched by *intersphinx*.
    """

    # matches a full name of an object, including the preceding module path with at
    # least one private submodule (starting with a "_") directly preceding the item
    # name
    __RE_PRIVATE_MODULE_AND_ITEM = re.compile(
        r"\b(?# we start with a word break so we match full words)"
        r"(?# public module path)(\w+(?:\.\w+)*?)"
        r"(?# private module path)((?:\._\w+)+)"
        r"\."
        r"(?# item name)(\w+(?![.\w]))"
    )

    def __init__(
        self,
        collapsible_submodules: Mapping[str, str],
        collapse_private_modules: bool = True,
    ):
        """
        :param collapsible_submodules: mapping from module paths to their public
            prefix, e.g., ``{"pandas.core.frame": "pandas"}``
        :param collapse_private_modules: if ``True``, collapse module sub-paths
            consisting of one or more protected modules (i.e. the module name starts
            with an underscore)
        """
        super().__init__()
        self._classes_visited: Set[type] = set()

        col = [
            self._make_substitution_pattern(old.replace(".", r"\."), new)
            for old, new in collapsible_submodules.items()
        ]

        self._intersphinx_collapsible_prefixes: List[Tuple[Pattern, str]] = col
        self._collapse_private_modules = collapse_private_modules

    @abstractmethod
    def _make_substitution_pattern(self, old: str, new: str) -> Tuple[Pattern, str]:
        # create the regex substitution rule given a raw match and replacement patterns
        pass

    def collapse_module_paths(self, line: Optional[str]) -> Optional[str]:
        """
        In the given line, replace all module paths with their collapsed version.

        :param line: the line in which to collapse module paths
        :return: the resulting line with collapsed module paths
        """
        if not line:
            return line

        if self._collapse_private_modules:
            line = self._collapse_private_module_paths(line)

        for expanded, collapsed in self._intersphinx_collapsible_prefixes:
            line = expanded.sub(collapsed, line)

        return line

    @staticmethod
    def _collapse_private_module_paths(line: str) -> str:
        for (
            public_module_path,  # e.g., "pytools.expression"
            private_module_path,  # e.g., "._expression"
            item_name,  # e.g., "Expression"
        ) in CollapseModulePaths.__RE_PRIVATE_MODULE_AND_ITEM.findall(line):
            module_path = public_module_path + private_module_path
            collapsed_path = public_module_path
            try:
                module = importlib.import_module(name=module_path)
                item = vars(module)[item_name]
                collapsed_path = item.__publicmodule__
            except KeyError:
                pass
            except AttributeError:
                pass
            except ModuleNotFoundError:
                pass

            line = line.replace(
                f"{module_path}.{item_name}", f"{collapsed_path}.{item_name}"
            )

        return line


@inheritdoc(match="[see superclass]")
class CollapseModulePathsInDocstring(CollapseModulePaths, AutodocProcessDocstring):
    """
    Replace private module paths in docstrings with their public prefix so that object
    references can be matched by *intersphinx*.
    """

    def process(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        options: object,
        lines: List[str],
    ) -> None:
        """[see superclass]"""

        for i, line in enumerate(lines):
            lines[i] = self.collapse_module_paths(line)

    def _make_substitution_pattern(self, old: str, new: str) -> Tuple[Pattern, str]:
        return re.compile(f"(`~?){old}"), f"\\1{new}"


@inheritdoc(match="[see superclass]")
class CollapseModulePathsInSignature(CollapseModulePaths, AutodocProcessSignature):
    """
    Replace private module paths in signatures with their public prefix so that object
    references can be matched by *intersphinx*.
    """

    def process(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        options: object,
        signature: Optional[str],
        return_annotation: Optional[str],
    ) -> Optional[Tuple[str, str]]:
        """[see superclass]"""
        if signature or return_annotation:
            return (
                self.collapse_module_paths(signature),
                self.collapse_module_paths(return_annotation),
            )

    def _make_substitution_pattern(self, old: str, new: str) -> Tuple[Pattern, str]:
        return re.compile(old), new


@inheritdoc(match="[see superclass]")
class SkipIndirectImports(AutodocSkipMember):
    """
    Skip members imported by a private package.
    """

    def process(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        skip: bool,
        options: object,
    ) -> Optional[bool]:
        """[see superclass]"""
        if not skip:
            if what == "module" and name.startswith("_"):
                log.info(f"skipping: {what}: {name}")
                return False


@inheritdoc(match="[see superclass]")
class Replace3rdPartyDoc(AutodocProcessDocstring):
    """
    Replace 3rd party docstrings with a reference to the 3rd party documentation.

    This is necessary for methods and attributes inherited from 3rd party packages,
    as these might use an incompatible format for docstrings.
    """

    __RE_ROOT_PACKAGE = re.compile(r"\w+(?=\.)")

    __RST_DIRECTIVE = {
        "module": "mod",
        "class": "class",
        "exception": "exception",
        "function": "func",
        "method": "meth",
        "attribute": "attr",
    }

    def process(
        self,
        app: Sphinx,
        what: str,
        name: str,
        obj: object,
        options: object,
        lines: List[str],
    ) -> None:
        """[see superclass]"""

        try:
            obj_module = obj.__module__
        except AttributeError:
            return

        name_root_package = self.__root_package(name)
        obj_root_package = self.__root_package(obj_module) if obj_module else None

        if name_root_package != obj_root_package:
            # replace 3rd party docstring with cross-reference

            directive = Replace3rdPartyDoc.__RST_DIRECTIVE.get(what, what)
            public_module = public_module_prefix(obj_module) if obj_module else ""

            del lines[:]
            lines.append(f"See :{directive}:`{public_module}.{obj.__qualname__}`")

    @staticmethod
    def __root_package(name: str) -> str:
        root_package_match = Replace3rdPartyDoc.__RE_ROOT_PACKAGE.match(name)
        return root_package_match[0] if root_package_match else ""


__tracker.validate()



"""
Base classes for dendrogram styles.
"""

import logging
from abc import ABCMeta, abstractmethod
from typing import Optional, Sequence

from matplotlib.axes import Axes
from matplotlib.colors import LogNorm

from pytools.api import AllTracker, inheritdoc
from pytools.viz import ColorbarMatplotStyle, DrawingStyle, MatplotStyle
from pytools.viz.color import ColorScheme
from pytools.viz.util import PercentageFormatter

log = logging.getLogger(__name__)

#
# Exported names
#

__all__ = ["DendrogramStyle", "DendrogramMatplotStyle"]


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Classes
#


class DendrogramStyle(DrawingStyle, metaclass=ABCMeta):
    """
    Base class for dendrogram drawing styles.
    """

    @abstractmethod
    def draw_leaf_names(
        self,
        *,
        names: Sequence[str],
    ) -> None:
        """
        Render the names for all leaves.

        :param names: the names of all leaves
        """
        pass

    @abstractmethod
    def draw_link_leg(
        self, bottom: float, top: float, leaf: float, weight: float, tree_height: float
    ) -> None:
        """
        Draw a "leg" connecting two levels of the linkage tree hierarchy.

        :param bottom: the height of the child node in the linkage tree
        :param top: the height of the parent node in the linkage tree
        :param leaf: the index of the leaf where the link leg should be drawn (may be
            a ``float``, indicating a position in between two leaves)
        :param weight: the weight of the child node
        :param tree_height: the total height of the linkage tree
        """
        pass

    @abstractmethod
    def draw_link_connector(
        self,
        bottom: float,
        top: float,
        first_leaf: int,
        n_leaves_left: int,
        n_leaves_right: int,
        weight: float,
        tree_height: float,
    ) -> None:
        """
        Draw a connector between two sub-trees and their parent node.

        :param bottom: the height (i.e. cluster distance) of the sub-trees
        :param top: the height of the parent node
        :param first_leaf: the index of the first leaf in the left sub-tree
        :param n_leaves_left: the number of leaves in the left sub-tree
        :param n_leaves_right: the number of leaves in the right sub-tree
        :param weight: the weight of the parent node
        :param tree_height: the total height of the linkage tree
        """
        pass


@inheritdoc(match="[see superclass]")
class DendrogramMatplotStyle(DendrogramStyle, ColorbarMatplotStyle, metaclass=ABCMeta):
    """
    Base class for `matplotlib` styles for dendrograms.

    Supports color maps to indicate feature importance on a logarithmic scale,
    and renders a color bar as a legend.
    """

    def __init__(
        self,
        *,
        ax: Optional[Axes] = None,
        colors: Optional[ColorScheme] = None,
        min_weight: float = 0.01,
    ) -> None:
        """
        :param min_weight: the minimum weight on the logarithmic feature importance
            color scale; must be greater than `0` and smaller than `1`
            (default: `0.01`, i.e., 1%)
        """
        if min_weight >= 1.0 or min_weight <= 0.0:
            raise ValueError("arg min_weight must be > 0.0 and < 1.0")

        percentage_formatter = PercentageFormatter()
        super().__init__(
            ax=ax,
            colors=colors,
            colormap_normalize=LogNorm(min_weight, 1),
            colorbar_major_formatter=percentage_formatter,
            colorbar_minor_formatter=percentage_formatter,
        )

    __init__.__doc__ = MatplotStyle.__init__.__doc__ + __init__.__doc__

    def draw_leaf_names(self, *, names: Sequence[str]) -> None:
        """[see superclass]"""
        ax = self.ax
        y_axis = ax.yaxis
        y_axis.set_ticks(ticks=range(len(names)))
        y_axis.set_ticklabels(ticklabels=names)

    def finalize_drawing(
        self,
        *,
        leaf_label: Optional[str] = None,
        distance_label: Optional[str] = None,
        weight_label: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Add labels to the axes of this drawing.

        :param leaf_label: the label for the leaf axis
        :param distance_label: the label for the distance axis
        :param weight_label: the label for the color bar, indicating weights
        """
        super().finalize_drawing(colorbar_label=weight_label, **kwargs)

        ax = self.ax

        # configure the axes
        ax.ticklabel_format(axis="x", scilimits=(-3, 3))
        if distance_label:
            ax.set_xlabel(distance_label)
        if leaf_label:
            ax.set_ylabel(leaf_label)


__tracker.validate()



"""
Implementation of :mod:`pytools.expression` and subpackages.
"""
import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Generic, Iterable, Tuple, TypeVar

from ...api import AllTracker, inheritdoc
from .. import Expression, HasExpressionRepr, make_expression
from ..operator import BinaryOperator, Operator

log = logging.getLogger(__name__)


#
# Exported names
#

__all__ = [
    "AtomicExpression",
    "SingletonExpression",
    "BracketPair",
    "BracketedExpression",
    "CollectionLiteral",
    "Operation",
    "PrefixExpression",
    "SimplePrefixExpression",
    "Invocation",
    "InfixExpression",
]

#
# Type variables
#

T = TypeVar("T")


#
# Ensure all symbols introduced below are included in __all__
#

__tracker = AllTracker(globals())


#
# Class definitions
#


#
# Atomic expressions
#


@inheritdoc(match="[see superclass]")
class AtomicExpression(Expression, Generic[T], metaclass=ABCMeta):
    """
    An atomic expression.

    Atomic expressions include literals and identifiers, and have no subexpressions.
    """

    @property
    def subexpressions_(self) -> Tuple[Expression, ...]:
        """
        The subexpressions of this expression
        (an empty tuple, as this expression is atomic).
        """
        return ()

    @property
    @abstractmethod
    def text_(self) -> str:
        """
        The text representing this atomic expression.
        """
        pass

    @property
    @abstractmethod
    def value_(self) -> T:
        """
        The underlying value of this atomic expression.
        """
        pass

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return BinaryOperator.MAX_PRECEDENCE

    def _eq_same_type(self, other: "AtomicExpression") -> bool:
        return self.value_ == other.value_

    def hash_(self) -> int:
        """[see superclass]"""
        return hash((type(self), self.value_))


#
# Complex expressions
#


class SingletonExpression(Expression, metaclass=ABCMeta):
    """
    An expression with a single subexpression.
    """

    @property
    @abstractmethod
    def subexpression_(self) -> Expression:
        """
        The subexpression of this expression.
        """

    @property
    def subexpressions_(self) -> Tuple[Expression, ...]:
        """
        A tuple with the this expression's subexpression as its only element.
        """
        return (self.subexpression_,)


#
# Bracketed expressions
#

# noinspection PyTypeHints,PyTypeChecker
_BracketPair = TypeVar("BracketPair", bound="BracketPair")


@inheritdoc(match="[see superclass]")
class BracketPair(HasExpressionRepr):
    """
    A pair of brackets.
    """

    #: A pair of round brackets.
    ROUND: _BracketPair = None
    #: A pair of square brackets.
    SQUARE: _BracketPair = None
    #: A pair of curly brackets.
    CURLY: _BracketPair = None
    #: A pair of angled brackets.
    ANGLED: _BracketPair = None

    #: The opening bracket.
    opening: str
    #: The closing bracket.
    closing: str

    def __init__(self, opening: str, closing: str) -> None:
        """
        :param opening: the opening bracket
        :param closing: the closing bracket
        """
        self.opening = opening
        self.closing = closing

    def to_expression(self) -> Expression:
        """[see superclass]"""
        from ..atomic import Id

        return Id(BracketPair.__name__)(opening=self.opening, closing=self.closing)


BracketPair.ROUND = BracketPair("(", ")")
BracketPair.SQUARE = BracketPair("[", "]")
BracketPair.CURLY = BracketPair("{", "}")
BracketPair.ANGLED = BracketPair("<", ">")


@inheritdoc(match="[see superclass]")
class BracketedExpression(SingletonExpression, metaclass=ABCMeta):
    """
    An expression surrounded by brackets.
    """

    def __init__(self, brackets: BracketPair, subexpression: Any) -> None:
        """
        :param brackets: the brackets enclosing this expression's subexpressions
        :param subexpression: the subexpression enclosed by brackets
        """
        super().__init__()
        self._brackets = brackets
        self._subexpression = make_expression(subexpression)

    @property
    def brackets_(self) -> BracketPair:
        """
        The brackets enclosing this expression's subexpressions.
        """
        return self._brackets

    @property
    def subexpression_(self) -> Expression:
        """
        The subexpression enclosed by the brackets.
        """
        return self._subexpression

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return Operator.MAX_PRECEDENCE

    def hash_(self) -> int:
        """[see superclass]"""
        return hash((type(self), self.brackets_, self.subexpression_.hash_()))

    def _eq_same_type(self, other: "BracketedExpression") -> bool:
        return self.brackets_ == other.brackets_ and self.subexpression_.eq_(
            other.subexpression_
        )


class CollectionLiteral(BracketedExpression):
    """
    A collection literal, e.g., a list, set, tuple, or dictionary.
    """

    def __init__(self, brackets: BracketPair, elements: Iterable[Any]) -> None:
        """
        :param brackets: the brackets enclosing the collection
        :param elements: the elements of the collection
        """

        from ..atomic import Epsilon
        from ..composite import BinaryOperation

        elements: Tuple[Expression, ...] = tuple(
            make_expression(element) for element in elements
        )

        subexpression: Expression
        if not elements:
            subexpression = Epsilon()
        elif len(elements) == 1:
            subexpression = elements[0]
        else:
            subexpression = BinaryOperation(BinaryOperator.COMMA, *elements)

        super().__init__(brackets, subexpression)

        self._elements = elements

    @property
    def subexpression_(self) -> Expression:
        """
        The subexpression representing the element(s) of this collection literal

        - an :class:`.Epsilon` expression for an empty collection
        - an arbitrary expression for a collection with a single element
        - a :class:`.BinaryOperation` with a :attr:`.BinaryOperator.COMMA`
          operator for collections with two or more elements
        """
        return super().subexpression_

    @property
    def elements_(self) -> Tuple[Expression]:
        """
        The expression(s) representing the elements of this expression.
        """
        return self._elements


#
# Operations
#


@inheritdoc(match="[see superclass]")
class Operation(Expression, metaclass=ABCMeta):
    """
    Abstract base class for operation expressions.
    """

    @property
    def subexpressions_(self) -> Tuple["Expression", ...]:
        """[see superclass]"""
        return self.operands_

    @property
    @abstractmethod
    def operator_(self) -> Operator:
        """
        The operator of this operation.
        """
        pass

    @property
    @abstractmethod
    def operands_(self) -> Tuple[Expression, ...]:
        """
        The operands of this operation; same as :attr:`subexpressions_`.
        """
        pass


#
# Prefix expressions
#


@inheritdoc(match="[see superclass]")
class PrefixExpression(Expression, metaclass=ABCMeta):
    """
    A prefix expression.

    Combines a prefix and a subexpression, optionally separated by one or more
    characters.
    """

    @property
    @abstractmethod
    def prefix_(self) -> Expression:
        """
        The prefix of this expression.
        """
        pass

    @property
    def separator_(self) -> str:
        """
        One or more characters separating the prefix and the subexpression.
        """
        return ""

    @property
    @abstractmethod
    def body_(self) -> Expression:
        """
        The body of this expression.
        """
        pass

    @property
    def subexpressions_(self) -> Tuple[Expression, ...]:
        """
        A tuple containing the prefix and the body of this expression.
        """
        return self.prefix_, self.body_

    def _eq_same_type(self, other: "PrefixExpression") -> bool:
        return self.prefix_.eq_(other.prefix_) and self.body_.eq_(other.body_)

    def hash_(self) -> int:
        """[see superclass]"""
        return hash(
            (type(self), self.prefix_.hash_(), self.separator_, self.body_.hash_())
        )


@inheritdoc(match="[see superclass]")
class SimplePrefixExpression(PrefixExpression, metaclass=ABCMeta):
    """
    An abstract base implementation of simple prefix expressions.
    """

    def __init__(self, prefix: Any, body: Any):
        """
        :param prefix: the prefix of the expression
        :param body: the body of the expression
        """
        super().__init__()
        self._prefix = make_expression(prefix)
        self._body = make_expression(body)

    @property
    def prefix_(self) -> Expression:
        """[see superclass]"""
        return self._prefix

    @property
    def body_(self) -> Expression:
        """[see superclass]"""
        return self._body


@inheritdoc(match="[see superclass]")
class Invocation(PrefixExpression):
    """
    An invocation in the shape of ``<expression>(<expression>)`` or
    ``<expression>[<expression>]``.
    """

    _PRECEDENCE = BinaryOperator.DOT.precedence

    def __init__(self, prefix: Any, brackets: BracketPair, args: Iterable[Any]):
        """
        :param prefix: the expression representing the object being invoked
        :param brackets: the brackets surrounding the invocation argument(s)
        :param args: the invocation argument(s); can also be an empty iterable
        """
        super().__init__()
        self._prefix = make_expression(prefix)
        self._invocation = CollectionLiteral(brackets=brackets, elements=args)

    @property
    def prefix_(self) -> Expression:
        """[see superclass]"""
        return self._prefix

    @property
    def body_(self) -> CollectionLiteral:
        """
        The expression representing the arguments enclosed by brackets.
        """
        return self._invocation

    @property
    def precedence_(self) -> int:
        """[see superclass]"""
        return Invocation._PRECEDENCE


#
# BinaryOperator expressions
#


@inheritdoc(match="[see superclass]")
class InfixExpression(Expression, metaclass=ABCMeta):
    """
    An infix expression, separating any number of subexpressions with an infix symbol.
    """

    @property
    @abstractmethod
    def infix_(self) -> BinaryOperator:
        """
        The infix used to separate this expression's subexpressions.
        """
        pass

    def hash_(self) -> int:
        """[see superclass]"""
        return hash(
            (
                type(self),
                self.infix_,
                *(subexpression.hash_() for subexpression in self.subexpressions_),
            )
        )

    def _eq_same_type(self, other: "InfixExpression") -> bool:
        return self.infix_ == other.infix_ and all(
            a.eq_(b) for a, b in zip(self.subexpressions_, other.subexpressions_)
        )


__tracker.validate()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pytools',
 'pytools.api',
 'pytools.data',
 'pytools.expression',
 'pytools.expression.atomic',
 'pytools.expression.base',
 'pytools.expression.composite',
 'pytools.expression.composite.base',
 'pytools.expression.formatter',
 'pytools.expression.operator',
 'pytools.fit',
 'pytools.meta',
 'pytools.parallelization',
 'pytools.sphinx',
 'pytools.text',
 'pytools.viz',
 'pytools.viz.color',
 'pytools.viz.dendrogram',
 'pytools.viz.dendrogram.base',
 'pytools.viz.distribution',
 'pytools.viz.distribution.base',
 'pytools.viz.matrix',
 'pytools.viz.matrix.base',
 'pytools.viz.util']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['joblib         >=0.13,<1.1',
 'matplotlib     >=3.0,<3.4',
 'numpy          >=1.16,<1.21',
 'pandas         >=0.24,<1.3',
 'scipy          >=1.2,<1.6',
 'typing_inspect >=0.4,<0.7']

extras_require = \
{'docs': ['sphinx == 3.4.*',
          'sphinx-autodoc-typehints == 1.11.*',
          'pydata-sphinx-theme == 0.4.*',
          'nbsphinx == 0.7.*',
          'jupyter >= 1.0',
          'docutils',
          'xlrd == 1.2.*',
          'm2r == 0.2.*'],
 'testing': ['pytest >= 5.2.*',
             'pytest-cov == 2.8.*',
             'flake8 == 3.8.*',
             'flake8-comprehensions == 3.2.*',
             'isort == 5.5.*']}

setup(name='gamma-pytools',
      version='1.0.4',
      description="A collection of Python extensions and tools used in BCG GAMMA's open-source libraries.",
      author='Boston Consulting Group (BCG)',
      author_email=None,
      url='https://github.com/BCG-Gamma/pytools',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6,<4',
     )



"""
Utilities for creating simulated data sets.
"""
from typing import Optional, Sequence

import numpy as np
import pandas as pd
from scipy.linalg import toeplitz

from ..api import AllTracker

__all__ = ["sim_data"]


__tracker = AllTracker(globals())


def sim_data(
    n: int = 100,
    intercept: float = -5,
    two_way_coef: Optional[float] = None,
    linear_vars: int = 10,
    linear_var_coef: Optional[Sequence[float]] = None,
    noise_vars: int = 0,
    corr_vars: int = 0,
    corr_type: str = "AR1",
    corr_value: float = 0,
    surg_err: float = 0.05,
    bin_var_p: float = 0,
    bin_coef: float = 0,
    outcome: str = "classification",
    regression_err: Optional[float] = None,
    seed_val: int = 4763546,
):
    """
    Simulate data for classification or regression that includes an interaction between
    two linear features, and some non-linear and linear features.

    Noise variables, correlated variables that are not predictive and surrogate features
    which are just derived from features that are predictive are also added.

    This function is for the most part a direct translation of the ``twoClassSim``
    function from the R package caret -- the option for an ordinal outcome and binary
    outcome mis-labelling were omitted. Full credit for the approach used for simulating
    binary classification data goes to the authors and contributors of caret
    [`Kuhn, M. (2008). Caret package. Journal of Statistical Software, 28(5).
    <https://rdrr.io/cran/caret/man/twoClassSim.html>`_]

    Key modifications compared to the *R* implementation:

    1.  The ordinal outcome option has not been translated
    2.  Mis-labelling of the binary outcome has not been translated
    3.  The addition of a linear feature that is a copy of another used in the linear
        predictor with a small amount of noise has been added to allow for the study
        of variable surrogacy
    4.  Option for a binary predictor and surrogate has been added
    5.  Toggle option for regression versus classification has been added
    6.  Arguments for the coefficients of primary predictors of interest has been added

    :param n: number of observations
    :param intercept: value for the intercept which can be modified to generate class
        imbalance
    :param two_way_coef: list of three coefficients: two linear terms and an
        interaction effect
    :param linear_vars: number of linear features
    :param linear_var_coef: an optional list of coefficients for linear features if
        the default is not desired
    :param noise_vars: number of unrelated independent noise features (do not
        contribute to the linear predictor)
    :param corr_vars: number of unrelated correlated noise features (do not contribute
        to the linear predictor)
    :param corr_type: type of correlation (exchangeable or auto-regressive) for
        correlated noise features
    :param corr_value: correlation for correlated noise features
    :param surg_err: degree of noise added to first linear predictor
    :param bin_var_p: prevalence for a binary feature to include in linear predictor
    :param bin_coef: coefficient for the impact of binary feature on linear predictor
    :param outcome: can be either classification for a binary outcome or regression
        for a continuous outcome
    :param regression_err: the error to be used in simulating a regression outcome
    :param seed_val: set a seed for reproducibility
    :return: data frame containing the simulated features and target for classification
    """

    # set seed
    np.random.seed(seed=seed_val)

    # add two correlated normal features for use in creating an interaction term in the
    # linear predictor
    sigma = np.array([[2, 1.3], [1.3, 2]])
    mu = [0, 0]
    tmp_data = pd.DataFrame(
        np.random.multivariate_normal(mu, sigma, size=n),
        columns=["TwoFactor1", "TwoFactor2"],
    )

    # add independent linear features that contribute to the linear predictor
    if linear_vars > 0:
        lin_cols = ["Linear" + str(x) for x in range(1, linear_vars + 1)]
        tmp_data = pd.concat(
            [
                tmp_data,
                pd.DataFrame(np.random.normal(size=(n, linear_vars)), columns=lin_cols),
            ],
            axis=1,
        )
    else:
        lin_cols = None

    # add non-linear features that contribute to the linear predictor
    tmp_data["Nonlinear1"] = pd.Series(np.random.uniform(low=-1.0, high=1.0, size=n))
    tmp_data = pd.concat(
        [
            tmp_data,
            pd.DataFrame(
                np.random.uniform(size=(n, 2)), columns=["Nonlinear2", "Nonlinear3"]
            ),
        ],
        axis=1,
    )

    # add independent noise features that do not contribute to the linear predictor
    if noise_vars > 0:
        noise_cols = ["Noise" + str(x) for x in range(1, noise_vars + 1)]
        tmp_data = pd.concat(
            [
                tmp_data,
                pd.DataFrame(
                    np.random.normal(size=(n, noise_vars)), columns=noise_cols
                ),
            ],
            axis=1,
        )

    # add correlated noise features that do not contribute to the linear predictor
    if corr_vars > 0:
        if corr_type == "exch":
            vc = corr_value * np.ones((corr_vars, corr_vars))
            np.fill_diagonal(vc, 1)

        elif corr_type == "AR1":
            vc_values = corr_value ** np.arange(corr_vars)
            vc = toeplitz(vc_values)

        else:
            raise ValueError(
                f'arg corr_type must be "exch" or "AR1", but got {repr(corr_type)}'
            )

        corr_cols = ["Corr" + str(x) for x in range(1, corr_vars + 1)]
        tmp_data = pd.concat(
            [
                tmp_data,
                pd.DataFrame(
                    np.random.multivariate_normal(np.zeros(corr_vars), vc, size=n),
                    columns=corr_cols,
                ),
            ],
            axis=1,
        )

    # add a surrogate linear feature that does not contribute to the linear predictor
    if linear_vars > 0:
        tmp_data["Linear1_prime"] = tmp_data["Linear1"] + np.random.normal(
            0, surg_err, size=n
        )

    # add a binary feature that contributes to the linear predictor
    if bin_var_p > 0:
        tmp_data["Binary1"] = np.where(np.random.uniform(size=n) <= bin_var_p, 0, 1)

    # generate linear predictor
    if two_way_coef is None:
        two_way_coef = [4, 4, 2]

    lp = (
        intercept
        - two_way_coef[0] * tmp_data.TwoFactor1
        + two_way_coef[1] * tmp_data.TwoFactor2
        + two_way_coef[2] * tmp_data.TwoFactor1 * tmp_data.TwoFactor2
        + tmp_data.Nonlinear1 ** 3
        + 2 * np.exp(-6 * (tmp_data.Nonlinear1 - 0.3) ** 2)
        + 2 * np.sin(np.pi * tmp_data.Nonlinear2 * tmp_data.Nonlinear3)
    )

    # add independent linear features to the linear predictor if required
    if linear_vars > 0:
        if linear_var_coef is None:
            lin_coef = np.linspace(linear_vars, 1, num=linear_vars) / 4
            neg_idx = list(range(1, linear_vars, 2))
            lin_coef[neg_idx] *= -1
            lp += tmp_data[lin_cols].dot(lin_coef)

        elif linear_var_coef is not None:
            if linear_vars != len(linear_var_coef):
                raise ValueError(
                    "User defined linear feature coefficient list must be of length "
                    f"{linear_vars}"
                )
            lp += tmp_data[lin_cols].dot(linear_var_coef)

    # add binary feature to the linear predictor if required
    if bin_var_p > 0:
        lp += bin_coef * tmp_data["Binary1"]
        tmp_data["Binary1_prime"] = 1 - tmp_data["Binary1"]

    # create classification outcome from linear predictor
    if outcome == "classification":

        # convert to a probability
        prob = 1 / (1 + np.exp(-lp))

        # generate target
        tmp_data["target"] = np.where(prob <= np.random.uniform(size=n), 0, 1)

    # create regression outcome
    elif outcome == "regression":

        # continuous outcome based on linear predictor
        tmp_data["target"] = np.random.normal(lp, regression_err, size=n)

    return tmp_data


__tracker.validate()



"""
Tests for module pytools.expression
"""

import logging

import pytest

from pytools.expression import Expression, freeze, make_expression
from pytools.expression.atomic import Id, Lit
from pytools.expression.composite import (
    BinaryOperation,
    Call,
    DictLiteral,
    Lambda,
    ListLiteral,
    SetLiteral,
    TupleLiteral,
    UnaryOperation,
)
from pytools.expression.formatter import PythonExpressionFormatter
from pytools.expression.formatter._python import TextualForm
from pytools.expression.operator import BinaryOperator, UnaryOperator

log = logging.getLogger(__name__)


def test_base_import() -> None:
    """
    Test if the expression base package can be imported without errors.
    """
    from pytools.expression import base

    assert base.AtomicExpression is not None


def test_expression_formatting() -> None:
    """
    Basic formatting tests
    """

    # expression 1

    e = Call(Id.f, (1 | Lit(2)) >> Lit("x") % Id.x, abc=-Lit(5))

    expr_1 = e * (e + e + e - e * e)
    form_1 = TextualForm.from_expression(expr_1)
    repr_1 = PythonExpressionFormatter(single_line=True).to_text(expr_1)
    assert len(form_1) == len(repr_1), f"length of {repr_1}"

    assert (
        str(expr_1)
        == """(
    f((1 | 2) >> 'x' % x, abc=-5)
    * (
        f((1 | 2) >> 'x' % x, abc=-5)
        + f((1 | 2) >> 'x' % x, abc=-5)
        + f((1 | 2) >> 'x' % x, abc=-5)
        - f((1 | 2) >> 'x' % x, abc=-5) * f((1 | 2) >> 'x' % x, abc=-5)
    )
)"""
    )

    # expression 2, generated with from_value

    expr_2 = make_expression([1, 2, {3: 4, 5: e}])
    assert str(expr_2) == "[1, 2, {3: 4, 5: f((1 | 2) >> 'x' % x, abc=-5)}]"

    # expression 3

    expr_3 = Id.g(param=(e, e + e, ~(e * e + e)))
    assert (
        str(expr_3)
        == """g(
    param=(
        f((1 | 2) >> 'x' % x, abc=-5),
        f((1 | 2) >> 'x' % x, abc=-5) + f((1 | 2) >> 'x' % x, abc=-5),
        ~(
            f((1 | 2) >> 'x' % x, abc=-5) * f((1 | 2) >> 'x' % x, abc=-5)
            + f((1 | 2) >> 'x' % x, abc=-5)
        )
    )
)"""
    )

    expr_4 = Lambda(body=e)
    assert repr(expr_4) == "lambda : f((1 | 2) >> 'x' % x, abc=-5)"

    expr_5 = Lambda(Id.x, body=e)(Lit(5))
    assert repr(expr_5) == "(lambda x: f((1 | 2) >> 'x' % x, abc=-5))(5)"

    expr_6 = Lambda(Id.x, Id.y, body=e)(Lit(5), 6)
    assert repr(expr_6) == "(lambda x, y: f((1 | 2) >> 'x' % x, abc=-5))(5, 6)"


def test_expression() -> None:
    lit_5 = Lit(5)
    lit_abc = Lit("abc")
    expressions = [
        (lit_5, "5"),
        (lit_abc, "'abc'"),
        (Id.xx, "xx"),
        (Call(Id("func"), lit_5, lit_abc), "func(5, 'abc')"),
        (ListLiteral(lit_5, lit_abc, Id.xx), "[5, 'abc', xx]"),
        (SetLiteral(lit_5, lit_abc, Id.xx), "{5, 'abc', xx}"),
        (TupleLiteral(lit_5, lit_abc, Id.xx), "(5, 'abc', xx)"),
        (DictLiteral(**{"5": lit_abc, "x": lit_5}), "{'5': 'abc', 'x': 5}"),
        (DictLiteral((lit_5, lit_abc), (Id.xx, lit_5)), "{5: 'abc', xx: 5}"),
        (BinaryOperation(BinaryOperator.ADD, lit_5, lit_abc, Id.xx), "5 + 'abc' + xx"),
        (Call(Id("func")), "func()"),
        (ListLiteral(), "[]"),
        (SetLiteral(), "{}"),
        (TupleLiteral(), "()"),
        (DictLiteral(), "{}"),
        (Id.xx.isalpha(), "xx.isalpha()"),
        (Id.xx[:], "xx[:]"),
        (Id.xx[::1], "xx[::1]"),
        (Id.xx[2::3, 1], "xx[2::3, 1]"),
        (Id.xx[4:], "xx[4:]"),
    ]

    for expression, expected_str in expressions:
        representation = TextualForm.from_expression(expression)
        assert len(representation) == len(expected_str), f"length of {representation}"
        assert str(representation) == expected_str
        assert len(
            PythonExpressionFormatter(single_line=True).to_text(expression)
        ) == len(expected_str)


def test_expression_setting() -> None:
    x = Id.x

    # we cannot assign by index
    with pytest.raises(TypeError):
        x[5] = 7

    # we cannot delete by index
    with pytest.raises(TypeError):
        del x[5]

    # we do not get an Attr expression for names with leading "_"
    with pytest.raises(AttributeError):
        x._private_a(3)

    # ... but we can assign private fields to the expression
    x._private_b = 3
    assert x._private_b == 3

    # we cannot assign values to public fields
    with pytest.raises(TypeError):
        x.public = 3


def test_comparison_expressions() -> None:
    x, y = Id.x, Id.y

    assert repr(x == y) == "x == y"
    assert repr(x != y) == "x != y"
    assert repr(x > y) == "x > y"
    assert repr(x >= y) == "x >= y"
    assert repr(x < y) == "x < y"
    assert repr(x <= y) == "x <= y"

    assert freeze(x) != freeze(y)
    assert not freeze(x) == freeze(y)

    a: Expression = make_expression([(x + (y * 3)), {freeze(y): x}])
    a_copy: Expression = make_expression([(x + (y * 3)), {freeze(y): x}])
    assert isinstance(a == a_copy, Expression)
    assert isinstance(freeze(a) == a_copy, bool)
    assert freeze(a) == freeze(a_copy)
    assert freeze(a) != a_copy
    assert freeze(a) != (a_copy + 1)


def test_expression_operators() -> None:
    a, b = Id.a, Id.b
    assert a + b == BinaryOperation(BinaryOperator.ADD, a, b)
    assert a - b == BinaryOperation(BinaryOperator.SUB, a, b)
    assert a * b == BinaryOperation(BinaryOperator.MUL, a, b)
    assert a @ b == BinaryOperation(BinaryOperator.MATMUL, a, b)
    assert a / b == BinaryOperation(BinaryOperator.DIV, a, b)
    assert a // b == BinaryOperation(BinaryOperator.FLOOR_DIV, a, b)
    assert a % b == BinaryOperation(BinaryOperator.MOD, a, b)
    assert a ** b == BinaryOperation(BinaryOperator.POW, a, b)
    assert a << b == BinaryOperation(BinaryOperator.LSHIFT, a, b)
    assert a >> b == BinaryOperation(BinaryOperator.RSHIFT, a, b)
    assert a & b == BinaryOperation(BinaryOperator.AND_BITWISE, a, b)
    assert a ^ b == BinaryOperation(BinaryOperator.XOR_BITWISE, a, b)
    assert a | b == BinaryOperation(BinaryOperator.OR_BITWISE, a, b)
    assert -a == UnaryOperation(UnaryOperator.NEG, operand=a)
    assert +a == UnaryOperation(UnaryOperator.POS, operand=a)
    assert ~a == UnaryOperation(UnaryOperator.INVERT, operand=a)
    assert (not a) == UnaryOperation(UnaryOperator.NOT, operand=a)


def test_operator_precedence() -> None:
    a, b, c = Id.a, Id.b, Id.c
    assert str(a + b + c) == "a + b + c"
    assert str((a + b) + c) == "a + b + c"
    assert str(a + (b + c)) == "a + (b + c)"
    assert str(a / b / c) == "a / b / c"
    assert str((a / b) / c) == "a / b / c"
    assert str(a / (b / c)) == "a / (b / c)"
    assert str(a * b + c) == "a * b + c"
    assert str((a * b) + c) == "a * b + c"
    assert str(a * (b + c)) == "a * (b + c)"
    assert str(a + b * c) == "a + b * c"
    assert str((a + b) * c) == "(a + b) * c"
    assert str(a + (b * c)) == "a + b * c"



#!/usr/bin/env python3
"""
Sphinx documentation build script
"""
import importlib
import importlib.util
import json
import os
import re
import shutil
import subprocess
import sys
from abc import ABCMeta, abstractmethod
from glob import glob
from tempfile import TemporaryDirectory
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Type, TypeVar
from weakref import ref

from packaging import version as pkg_version

cwd = os.getcwd()

# Sphinx commands
CMD_SPHINX_BUILD = "sphinx-build"
CMD_SPHINX_AUTOGEN = "sphinx-autogen"

# File paths
DIR_MAKE_BASE = os.path.dirname(os.path.realpath(__file__))
DIR_REPO_ROOT = os.path.realpath(os.path.join(os.getcwd(), os.pardir))
DIR_REPO_PARENT = os.path.realpath(os.path.join(DIR_REPO_ROOT, os.pardir))
FACET_PROJECT = os.path.split(os.path.realpath(DIR_REPO_ROOT))[1]
DIR_PACKAGE_SRC = os.path.join(DIR_REPO_ROOT, "src")
DIR_DOCS = os.path.join(DIR_REPO_ROOT, "docs")
DIR_SPHINX_SOURCE = os.path.join(cwd, "source")
DIR_SPHINX_AUX = os.path.join(cwd, "auxiliary")
DIR_SPHINX_API_GENERATED = os.path.join(DIR_SPHINX_SOURCE, "apidoc")
DIR_SPHINX_GET_STARTED_GENERATED = os.path.join(DIR_SPHINX_SOURCE, "getting_started")
DIR_SPHINX_BUILD = os.path.join(cwd, "build")
DIR_SPHINX_BUILD_HTML = os.path.join(DIR_SPHINX_BUILD, "html")
DIR_SPHINX_TEMPLATES = os.path.join(DIR_SPHINX_SOURCE, "_templates")
DIR_SPHINX_SOURCE_BASE = os.path.join(DIR_MAKE_BASE, os.pardir, "source")
DIR_SPHINX_TEMPLATES_BASE = os.path.join(DIR_SPHINX_SOURCE_BASE, "_templates")
DIR_SPHINX_AUTOSUMMARY_TEMPLATE = os.path.join(DIR_SPHINX_TEMPLATES, "autosummary.rst")
DIR_SPHINX_TUTORIAL = os.path.join(DIR_SPHINX_SOURCE, "tutorial")
DIR_SPHINX_SOURCE_STATIC_BASE = os.path.join(DIR_SPHINX_SOURCE_BASE, "_static_base")
JS_VERSIONS_FILE = os.path.join(DIR_SPHINX_SOURCE_STATIC_BASE, "js", "versions.js")
DIR_ALL_DOCS_VERSIONS = os.path.join(DIR_SPHINX_BUILD, "docs-version")

# Environment variables
# noinspection SpellCheckingInspection
ENV_PYTHON_PATH = "PYTHONPATH"

T = TypeVar("T")


class CommandMeta(ABCMeta):
    """
    Meta-class for command classes.

    Subsequent instantiations of a command class return the identical object.
    """

    def __init__(cls, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        cls.__instance_ref: Optional[ref] = None

    def __call__(cls: Type[T], *args, **kwargs: Any) -> T:
        """
        Return the existing command instance, or create a new one if none exists yet.

        :return: the command instance
        """
        if args or kwargs:
            raise ValueError("command classes may not take any arguments")

        cls: CommandMeta

        if cls.__instance_ref:
            obj = cls.__instance_ref()
            if obj is not None:
                return obj

        instance = super(CommandMeta, cls).__call__()
        cls.__instance_ref = ref(instance)
        return instance


class Command(metaclass=CommandMeta):
    """ Defines an available command that can be launched from this module."""

    __RE_CAMEL_TO_SNAKE = re.compile(r"(?<!^)(?=[A-Z])")

    def __init__(self) -> None:
        self.name = self.__RE_CAMEL_TO_SNAKE.sub("_", type(self).__name__).lower()

    @abstractmethod
    def get_description(self) -> str:
        pass

    @abstractmethod
    def get_dependencies(self) -> Tuple["Command", ...]:
        pass

    def get_prerequisites(self) -> Iterable["Command"]:
        dependencies_extended: List[Command] = []

        for dependency in self.get_dependencies():
            dependencies_inherited = dependency.get_dependencies()
            if self in dependencies_inherited:
                raise ValueError(
                    f"circular dependency: {dependency.name} " f"depends on {self.name}"
                )
            dependencies_extended.extend(
                dependency
                for dependency in dependencies_inherited
                if dependency not in dependencies_extended
            )
            dependencies_extended.append(dependency)

        return dependencies_extended

    def run(self) -> None:
        print(f"Running command {self.name} – {self.get_description()}")
        self._run()

    @abstractmethod
    def _run(self) -> None:
        pass


#
# commands
#


class Clean(Command):
    def get_description(self) -> str:
        return "remove Sphinx build output"

    def get_dependencies(self) -> Tuple[Command, ...]:
        return ()

    def _run(self) -> None:
        if os.path.exists(DIR_SPHINX_BUILD):
            shutil.rmtree(path=DIR_SPHINX_BUILD)
        if os.path.exists(DIR_SPHINX_API_GENERATED):
            shutil.rmtree(path=DIR_SPHINX_API_GENERATED)
        if os.path.exists(DIR_SPHINX_GET_STARTED_GENERATED):
            shutil.rmtree(path=DIR_SPHINX_GET_STARTED_GENERATED)


class ApiDoc(Command):
    def get_description(self) -> str:
        return "generate Sphinx API documentation from sources"

    def get_dependencies(self) -> Tuple[Command, ...]:
        # noinspection PyRedundantParentheses
        return (Clean(),)

    def _run(self) -> None:
        packages = [
            package for package in os.listdir(DIR_PACKAGE_SRC) if package[:1].isalnum()
        ]
        print(f"Generating api documentation for {', '.join(packages)}")

        package_lines = "\n   ".join(packages)
        # noinspection SpellCheckingInspection
        autosummary_rst = f""".. autosummary::
   :toctree: ../apidoc
   :template: custom-module-template.rst
   :recursive:

   {package_lines}
"""

        os.makedirs(os.path.dirname(DIR_SPHINX_AUTOSUMMARY_TEMPLATE), exist_ok=True)
        with open(DIR_SPHINX_AUTOSUMMARY_TEMPLATE, "wt") as f:
            f.writelines(autosummary_rst)
        autogen_options = " ".join(
            [
                # template path
                "-t",
                quote_path(DIR_SPHINX_TEMPLATES_BASE),
                # include imports
                "-i",
                # the autosummary source file
                quote_path(DIR_SPHINX_AUTOSUMMARY_TEMPLATE),
            ]
        )

        subprocess.run(
            args=f"{CMD_SPHINX_AUTOGEN} {autogen_options}",
            shell=True,
            check=True,
        )


class GettingStartedDoc(Command):
    def get_description(self) -> str:
        return "generate getting started documentation from sources"

    def get_dependencies(self) -> Tuple[Command, ...]:
        # noinspection PyRedundantParentheses
        return (Clean(),)

    def _run(self) -> None:

        # make dir if it does not exist
        os.makedirs(DIR_SPHINX_GET_STARTED_GENERATED, exist_ok=True)

        # open the rst readme file
        with open(os.path.join(DIR_REPO_ROOT, "README.rst"), "r") as file:
            readme_data = file.read()

        # modify links (step back needed as build will add sub directory to paths)
        readme_data = readme_data.replace("sphinx/source/", "../")
        readme_data = readme_data.replace("sphinx/auxiliary/", "../")
        readme_data = re.sub(
            r"\.\. Begin-Badges.*?\.\. End-Badges", "", readme_data, flags=re.S
        )

        with open(os.path.join(DIR_SPHINX_SOURCE, "release_notes.rst"), "w") as dst:
            dst.write(".. _release-notes:\n\n")
            with open(os.path.join(DIR_REPO_ROOT, "RELEASE_NOTES.rst"), "r") as src:
                dst.write(src.read())

        # create a new getting_started.rst that combines the header from templates and
        # adds an include for the README
        with open(
            os.path.join(DIR_SPHINX_TEMPLATES_BASE, "getting-started-header.rst"), "r"
        ) as file:
            template_data = file.read()

        with open(
            os.path.join(DIR_SPHINX_GET_STARTED_GENERATED, "getting_started.rst"),
            "wt",
        ) as file:
            file.writelines(template_data)
            file.writelines(readme_data)


class FetchPkgVersions(Command):
    def get_description(self) -> str:
        return "fetch available package versions with docs"

    def get_dependencies(self) -> Tuple[Command, ...]:
        return ()

    def _run(self) -> None:
        versions = Versions()

        version_data = {
            "current": str(versions.latest_stable_version),
            "non_rc": list(map(str, versions.version_tags_stable)),
            "all": list(map(str, versions.version_tags)),
        }

        version_data_as_js = (
            f"const DOCS_VERSIONS = {json.dumps(version_data, indent=4,)}"
        )

        with open(JS_VERSIONS_FILE, "wt") as f:
            f.write(version_data_as_js)

        print(f"Version data written into: {JS_VERSIONS_FILE}")


class PrepareDocsDeployment(Command):
    def get_description(self) -> str:
        return "update versions of rendered documentation"

    def get_dependencies(self) -> Tuple[Command, ...]:
        return ()

    def _run(self) -> None:
        if not is_azure_build():
            raise RuntimeError("only implemented for Azure Pipelines")

        # get the current version of the package
        current_version: pkg_version.Version = get_package_version()

        # copy new the docs version to the deployment path
        if current_version == Versions().latest_stable_version:
            # only copy to deployment path if our version is the latest stable release

            # remove docs build currently deployed, except for the docs versions folder
            if os.path.exists(os.path.join(DIR_DOCS, "docs-version")):
                with TemporaryDirectory() as DIR_TMP:
                    shutil.move(src=os.path.join(DIR_DOCS, "docs-version"), dst=DIR_TMP)
                    shutil.rmtree(path=DIR_DOCS)
                    os.makedirs(DIR_DOCS)
                    shutil.move(src=os.path.join(DIR_TMP, "docs-version"), dst=DIR_DOCS)
            else:
                os.makedirs(DIR_DOCS, exist_ok=True)

            # copy new docs version to deployment path
            shutil.copytree(src=DIR_SPHINX_BUILD_HTML, dst=DIR_DOCS, dirs_exist_ok=True)

        # get current version of package in the form of folder/URL name (e.g., "1-0-0")
        current_version_path = os.path.join(
            DIR_DOCS, "docs-version", version_string_to_url(current_version)
        )
        # update latest version in docs history
        if os.path.exists(current_version_path):
            shutil.rmtree(path=current_version_path)
        shutil.copytree(
            src=DIR_ALL_DOCS_VERSIONS,
            dst=os.path.join(DIR_DOCS, "docs-version"),
            dirs_exist_ok=True,
        )

        # Replace all docs version lists with the most up-to-date to have all versions
        # accessible also from older versions
        new_versions_js = os.path.join(DIR_DOCS, "_static", "js", "versions.js")
        for d in glob(os.path.join(DIR_DOCS, "docs-version", "*", "")):
            old_versions_js = os.path.join(d, "_static", "js", "versions.js")
            shutil.copyfile(src=new_versions_js, dst=old_versions_js)

        # remove .buildinfo which interferes with GitHub Pages build
        if os.path.exists(os.path.join(DIR_DOCS, ".buildinfo")):
            os.remove(os.path.join(DIR_DOCS, ".buildinfo"))

        # create empty file to signal that no GitHub auto-rendering is required
        # noinspection SpellCheckingInspection
        open(os.path.join(DIR_DOCS, ".nojekyll"), "a").close()

        print("Docs moved to ./docs and historic versions updated")


class Html(Command):
    def get_description(self) -> str:
        return "build Sphinx docs as HTML"

    def get_dependencies(self) -> Tuple[Command, ...]:
        return Clean(), FetchPkgVersions(), ApiDoc(), GettingStartedDoc()

    def _run(self) -> None:

        check_sphinx_version()

        os.makedirs(DIR_SPHINX_BUILD, exist_ok=True)

        sphinx_html_opts = [
            "-M html",
            quote_path(DIR_SPHINX_SOURCE),
            quote_path(DIR_SPHINX_BUILD),
        ]

        subprocess.run(
            args=f"{CMD_SPHINX_BUILD} {' '.join(sphinx_html_opts)}",
            shell=True,
            check=True,
        )

        # create interactive versions of all notebooks
        sys.path.append(DIR_MAKE_BASE)

        # create copy of this build for the docs archive
        version_built: pkg_version.Version = get_package_version()
        dir_path_this_build = os.path.join(
            DIR_ALL_DOCS_VERSIONS, version_string_to_url(version_built)
        )

        os.makedirs(DIR_ALL_DOCS_VERSIONS, exist_ok=True)
        if os.path.exists(dir_path_this_build):
            shutil.rmtree(dir_path_this_build)

        shutil.copytree(
            src=DIR_SPHINX_BUILD_HTML,
            dst=dir_path_this_build,
        )

        if not is_azure_build():
            shutil.move(src=DIR_ALL_DOCS_VERSIONS, dst=DIR_SPHINX_BUILD_HTML)

        # empty versions file to blank template
        with open(JS_VERSIONS_FILE, "wt") as f:
            f.write("")


class Help(Command):
    def get_description(self) -> str:
        return "print this help message"

    def get_dependencies(self) -> Tuple[Command, ...]:
        return ()

    def _run(self) -> None:
        print_usage()


class Versions:
    """
    Helper class that lists all versions that have already been released.
    """

    INITIAL_VERSION_TAG = pkg_version.parse("1.0.1")

    def __init__(self) -> None:
        os.makedirs(DIR_SPHINX_BUILD, exist_ok=True)
        start_from_version_tag: pkg_version.Version = Versions.INITIAL_VERSION_TAG
        sp = subprocess.run(
            args='git tag -l "*.*.*"', shell=True, check=True, stdout=subprocess.PIPE
        )
        version_tags: List[pkg_version.Version] = [
            version_tag
            for version_tag in (
                pkg_version.parse(version_string)
                for version_string in sp.stdout.decode("UTF-8").split("\n")
                if version_string
            )
            if version_tag >= start_from_version_tag
        ]

        # add version currently build into version_tags
        version_built: pkg_version.Version = get_package_version()

        if version_built not in version_tags:
            version_tags.append(version_built)

        version_tags.sort()
        version_tags.reverse()
        version_tags_stable: List[pkg_version.Version] = [
            vt for vt in version_tags if not (vt.is_prerelease or vt.is_devrelease)
        ]
        latest_stable_version: pkg_version.Version = version_tags_stable[0]

        print(f"Found versions: {', '.join(map(str, version_tags))}")
        print("Latest stable version: ", latest_stable_version)

        self.version_tags = version_tags
        self.version_tags_stable = version_tags_stable
        self.latest_stable_version = latest_stable_version


def make(*, modules: List[str]) -> None:
    """
    Run this make script with the given arguments.
    """
    if len(sys.argv) < 2:
        print_usage()

    commands_passed = sys.argv[1:]

    unknown_commands = set(commands_passed) - available_commands.keys()

    if unknown_commands:
        print(f"Unknown build commands: {' '.join(unknown_commands)}\n")
        print_usage()
        exit(1)

    # set up the Python path
    module_paths = [
        os.path.abspath(os.path.join(DIR_REPO_PARENT, module, "src"))
        for module in modules
    ]
    if ENV_PYTHON_PATH in os.environ:
        module_paths.append(os.environ[ENV_PYTHON_PATH])
    os.environ[ENV_PYTHON_PATH] = os.pathsep.join(module_paths)

    # run all given commands:
    executed_commands: Set[Command] = set()

    for next_command_name in commands_passed:

        next_command: Command = available_commands[next_command_name]

        for prerequisite_command in next_command.get_prerequisites():

            if prerequisite_command not in executed_commands:
                prerequisite_command.run()
                executed_commands.add(prerequisite_command)

        next_command.run()
        executed_commands.add(next_command)


def quote_path(path: str) -> str:
    """
    Quote a file path if it contains whitespace.
    """
    if " " in path or "\t" in path:
        return f'"{path}"'
    else:
        return path


def get_package_version() -> pkg_version.Version:
    """
    Retrieve the package version for the project from __init__ or _version
    """
    project_src = os.path.abspath(os.path.join(DIR_REPO_ROOT, "src"))

    if FACET_PROJECT in ("sklearndf", "flow"):
        # for sklearndf and flow __init__ can't be trivially imported due to import
        # dependencies. Load the version as defined in FACET_PROJECT._version module
        spec = importlib.util.spec_from_file_location(
            "_version", os.path.join(project_src, FACET_PROJECT, "_version.py")
        )
    else:
        # pytools/facet: retrieve version from __init__.py
        spec = importlib.util.spec_from_file_location(
            "_version", os.path.join(project_src, FACET_PROJECT, "__init__.py")
        )

    version_module = importlib.util.module_from_spec(spec)
    # noinspection PyUnresolvedReferences
    spec.loader.exec_module(version_module)
    # noinspection PyUnresolvedReferences
    return pkg_version.parse(version_module.__version__)


def is_azure_build() -> bool:
    """
    Check if this is an Azure DevOps pipelines build
    """
    return "BUILD_REASON" in os.environ


def version_string_to_url(version: pkg_version.Version) -> str:
    """
    Make a Python package version string safe for URLs/folders.

    Our convention is to only replace all dots with dashes.
    """
    return str(version).replace(".", "-")


def check_sphinx_version() -> None:
    import sphinx

    sphinx_version = pkg_version.parse(sphinx.__version__)
    if sphinx_version < pkg_version.parse("3.2.1"):
        raise RuntimeError("please upgrade sphinx to version 3.2.1 or newer")


def print_usage() -> None:
    usage = """Sphinx documentation build script
=================================

Available program arguments:
"""
    usage += "\n".join(
        f"\t{name} – {command.get_description()}"
        for name, command in available_commands.items()
    )
    print(usage)


available_commands: Dict[str, Command] = {
    cmd.name: cmd
    for cmd in (
        Clean(),
        ApiDoc(),
        GettingStartedDoc(),
        Html(),
        Help(),
        FetchPkgVersions(),
        PrepareDocsDeployment(),
    )
}



"""
Configuration file for the Sphinx documentation builder.

For a full list of options see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.realpath to make it absolute, like shown here.
#

import logging
import os
import shutil
import sys
from typing import Any, Dict, Iterable, Optional

from sphinx.application import Sphinx

from pytools.sphinx import Replace3rdPartyDoc

logging.basicConfig(level=logging.INFO)
_log = logging.getLogger(name=__name__)

# this is the directory that contains all required repos
_dir_conf_base = os.path.dirname(os.path.realpath(__file__))
_dir_repo_root = os.path.normpath(
    os.path.join(_dir_conf_base, os.pardir, os.pardir, os.pardir)
)
_dir_sphinx = os.path.abspath(os.getcwd())


# noinspection PyShadowingNames
def set_config(
    globals_: Dict[str, Any],
    *,
    project: str,
    modules: Iterable[str],
    html_logo: Optional[str] = None,
) -> None:
    """
    Add required modules to the python path, and set custom configuration options
    """

    globals_["project"] = project

    if html_logo:
        globals_["html_logo"] = html_logo
        globals_["latex_logo"] = html_logo

    modules = set(modules) | {"pytools"}
    for module in modules:
        module_path = os.path.normpath(os.path.join(_dir_repo_root, module, "src"))
        if module_path not in sys.path:
            # noinspection PyUnboundLocalVariable
            sys.path.insert(0, module_path)
            _log.info(f"added `{module_path}` to python paths")

    # Update global variables
    globals_.update(
        (k, v) for k, v in globals().items() if not (k.startswith("_") or k in globals_)
    )

    globals_.update(
        (k, v) for k, v in globals().items() if not (k.startswith("_") or k in globals_)
    )


_log.info(f"sys.path = {sys.path}")

# -- Project information -----------------------------------------------------

project = "pytools"
copyright = "2021, Boston Consulting Group (BCG)"
author = "FACET Team"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
]

# -- Options for autodoc / autosummary -------------------------------------------------

# prevent numpydoc from interfering with autosummary
numpydoc_show_class_members = False

# generate autosummary even if no references
autosummary_generate = False

# always overwrite generated auto summaries with newly generated versions
autosummary_generate_overwrite = True

autosummary_imported_members = True

autodoc_default_options = {
    "ignore-module-all": False,
    "inherited-members": True,
    "imported-members": False,
    "show-inheritance": False,
    "member-order": "groupwise",
}

nbsphinx_allow_errors = True

# add intersphinx mapping
intersphinx_mapping = {
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "pd": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "matplotlib": ("https://matplotlib.org", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "np": ("https://numpy.org/doc/stable", None),
    "python": ("https://docs.python.org/3.6", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "sklearn": ("https://scikit-learn.org/stable", None),
    "shap": ("https://shap.readthedocs.io/en/stable", None),
    "joblib": ("https://joblib.readthedocs.io/en/latest", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "lightgbm": ("https://lightgbm.readthedocs.io/en/latest/", None),
    "pytools": ("https://bcg-gamma.github.io/pytools/", None),
    "sklearndf": ("https://bcg-gamma.github.io/sklearndf/", None),
    "facet": ("https://bcg-gamma.github.io/facet/", None),
}

intersphinx_collapsible_submodules = {
    "pandas.core.frame": "pandas",
    "pandas.core.series": "pandas",
    "pandas.core.indexes.base": "pandas",
    "pandas.core.indexes.multi": "pandas",
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

source_suffix = [".rst", ".md", ".ipynb"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["*/.ipynb_checkpoints/*"]

# -- Options for sphinx_autodoc_typehints ----------------------------------------------
set_type_checking_flag = False
typehints_fully_qualified = False
always_document_param_types = False

# -- Options for Math output -----------------------------------------------------------

imgmath_image_format = "svg"
imgmath_use_preview = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_logo = "_static/gamma_logo.jpg"

# Class documentation to include docstrings both global to the class, and from __init__
autoclass_content = "both"


def _get_package_version() -> str:
    """
    Get the package version for the project being built.
    :return: string with Python package version
    """
    # NOTE: sphinx-make changes the CWD into sphinx/source
    #       while FACET's make_base expects a CWD <project>/src
    #       hence: save current CWD, navigate to <project>/src, then go back
    cwd = os.getcwd()
    os.chdir(os.path.join(cwd, os.pardir, os.pardir, "src"))
    import make_base

    version_ = str(make_base.get_package_version())
    os.chdir(cwd)
    return version_


version = _get_package_version()
# -- End of options section ------------------------------------------------------------


def setup(app: Sphinx) -> None:
    """
    Add event handlers to the Sphinx application object
    :param app: the Sphinx application object
    """

    from pytools.sphinx import (
        AddInheritance,
        CollapseModulePathsInDocstring,
        CollapseModulePathsInSignature,
        SkipIndirectImports,
    )

    AddInheritance(collapsible_submodules=intersphinx_collapsible_submodules).connect(
        app=app
    )

    CollapseModulePathsInDocstring(
        collapsible_submodules=intersphinx_collapsible_submodules
    ).connect(app=app, priority=100000)

    CollapseModulePathsInSignature(
        collapsible_submodules=intersphinx_collapsible_submodules
    ).connect(app=app, priority=100000)

    SkipIndirectImports().connect(app=app)

    Replace3rdPartyDoc().connect(app=app)

    _add_custom_css_and_js(app=app)


def _add_custom_css_and_js(app: Sphinx):
    # add custom css and js files, and copy them to the build/html/_static folder
    css_rel_paths = (os.path.join("css", "gamma.css"),)
    js_rel_paths = (os.path.join("js", "gamma.js"), os.path.join("js", "versions.js"))

    for css in css_rel_paths:
        app.add_css_file(filename=css)

    for js in js_rel_paths:
        app.add_js_file(filename=js)

    src_root = os.path.normpath(
        os.path.join(_dir_conf_base, os.pardir, "source", "_static_base")
    )
    dst_html = os.path.normpath(
        os.path.join(_dir_sphinx, os.pardir, "build", "html", "_static")
    )
    for rel_path in css_rel_paths + js_rel_paths:
        dst_dir = os.path.join(dst_html, os.path.dirname(rel_path))
        os.makedirs(dst_dir, exist_ok=True)
        src_file = os.path.join(src_root, rel_path)
        print(f"copying {src_file} to {dst_dir}")
        shutil.copy(src=src_file, dst=dst_dir)
